"""
This module implements the methods for calculating the plasma wakefields
using the 2D r-z reduced model from P. Baxevanis and G. Stupakov.

See https://journals.aps.org/prab/abstract/10.1103/PhysRevAccelBeams.21.071301
for the full details about this model.
"""

import numpy as np
import scipy.constants as ct
import aptools.plasma_accel.general_equations as ge

from wake_t.particles.deposition import deposit_3d_distribution
from .deposition import deposit_plasma_particles
from wake_t.particles.interpolation import gather_sources_qs_baxevanis
from wake_t.utilities.other import radial_gradient
from .plasma_push.rk4 import evolve_plasma_rk4
from .plasma_push.ab5 import evolve_plasma_ab5
from .psi_and_derivatives import (
    calculate_psi,
    calculate_psi_and_derivatives_at_particles,
)
from .b_theta import calculate_b_theta, calculate_b_theta_at_particles
from .plasma_particles import PlasmaParticles
from wake_t.utilities.numba import njit_serial


def calculate_wakefields(
    laser_a2,
    bunches,
    r_max,
    xi_min,
    xi_max,
    n_r,
    n_xi,
    ppc,
    n_p,
    r_max_plasma=None,
    parabolic_coefficient=0.0,
    p_shape="cubic",
    max_gamma=10.0,
    plasma_pusher="rk4",
    fld_arrays=[],
):
    """
    Calculate the plasma wakefields generated by the given laser pulse and
    electron beam in the specified grid points.

    Parameters
    ----------
    laser_a2 : ndarray
        A (nz x nr) array containing the square of the laser envelope.
    beam_part : list
        List of numpy arrays containing the spatial coordinates and charge of
        all beam particles, i.e [x, y, xi, q].
    r_max : float
        Maximum radial position up to which plasma wakefield will be
        calculated.
    xi_min : float
        Minimum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    xi_max : float
        Maximum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    n_r : int
        Number of grid elements along r in which to calculate the wakefields.
    n_xi : int
        Number of grid elements along xi in which to calculate the wakefields.
    ppc : int (optional)
        Number of plasma particles per 1d cell along the radial direction.
    n_p : float
        Plasma density in units of m^{-3}.
    r_max_plasma : float
        Maximum radial extension of the plasma column. If `None`, the plasma
        extends up to the `r_max` boundary of the simulation box.
    parabolic_coefficient : float
        The coefficient for the transverse parabolic density profile. The
        radial density distribution is calculated as
        `n_r = n_p * (1 + parabolic_coefficient * r**2)`, where `n_p` is the
        local on-axis plasma density.
    p_shape : str
        Particle shape to be used for the beam charge deposition. Possible
        values are 'linear' or 'cubic'.
    max_gamma : float
        Plasma particles whose `gamma` exceeds `max_gamma` are considered to
        violate the quasistatic condition and are put at rest (i.e.,
        `gamma=1.`, `pr=pz=0.`).
    plasma_pusher : str
        Numerical pusher for the plasma particles. Possible values are `'rk4'`
        and `'ab5'`.

    """
    rho, chi, E_r, E_z, B_t, xi_fld, r_fld = fld_arrays

    s_d = ge.plasma_skin_depth(n_p * 1e-6)
    r_max = r_max / s_d
    xi_min = xi_min / s_d
    xi_max = xi_max / s_d
    dr = r_max / n_r
    dxi = (xi_max - xi_min) / (n_xi - 1)
    parabolic_coefficient = parabolic_coefficient * s_d**2

    # Maximum radial extent of the plasma.
    if r_max_plasma is None:
        r_max_plasma = r_max
    else:
        r_max_plasma = r_max_plasma / s_d

    # Initialize plasma particles.
    pp = PlasmaParticles(
        r_max, r_max_plasma, parabolic_coefficient, dr, ppc, plasma_pusher
    )
    pp.initialize()

    # Initialize field arrays, including guard cells.
    a2 = np.zeros((n_xi + 4, n_r + 4))
    nabla_a2 = np.zeros((n_xi + 4, n_r + 4))
    psi = np.zeros((n_xi + 4, n_r + 4))
    W_r = np.zeros((n_xi + 4, n_r + 4))
    b_t_bar = np.zeros((n_xi + 4, n_r + 4))

    # Field node coordinates.
    r_fld = r_fld / s_d
    xi_fld = xi_fld / s_d

    # Laser source.
    a2[2:-2, 2:-2] = laser_a2
    nabla_a2[2:-2, 2:-2] = radial_gradient(laser_a2, dr)

    # Beam source. This code is needed while no proper support particle
    # beams as input is implemented.
    b_t_beam = np.zeros((n_xi + 4, n_r + 4))
    for bunch in bunches:
        calculate_beam_source(
            bunch, n_p, n_r, n_xi, r_fld[0], xi_fld[0], dr, dxi, p_shape, b_t_beam
        )

    # Evolve plasma from right to left and calculate psi, b_t_bar, rho and
    # chi on a grid.
    evolve_plasma_and_calculate_fields(
        pp,
        a2,
        nabla_a2,
        b_t_beam,
        psi,
        b_t_bar,
        rho,
        chi,
        xi_fld,
        r_fld,
        dxi,
        dr,
        n_xi,
        n_r,
        max_gamma,
        p_shape,
        plasma_pusher,
    )

    # Calculate derived fields (E_z, W_r, and E_r).
    E_0 = ge.plasma_cold_non_relativisct_wave_breaking_field(n_p * 1e-6)
    dxi_psi, dr_psi = np.gradient(psi[2:-2, 2:-2], dxi, dr, edge_order=2)
    E_z[2:-2, 2:-2] = -dxi_psi * E_0
    W_r[2:-2, 2:-2] = -dr_psi * E_0
    B_t[:] = (b_t_bar + b_t_beam) * E_0 / ct.c
    E_r[:] = W_r + B_t * ct.c


def evolve_plasma_and_calculate_fields(
    pp,
    a2,
    nabla_a2,
    b_t_0,
    psi,
    b_t_bar,
    rho,
    chi,
    xi_fld,
    r_fld,
    dxi,
    dr,
    n_xi,
    n_r,
    max_gamma,
    p_shape,
    plasma_pusher,
):
    """Evolve plasma column from right to left and calculate plasma fields.

    Parameters
    ----------
    pp : PlasmaParticles
        The column of plasma particles to be evolved.
    a2, nabla_a2, b_t_0 : ndarray
        Arrays containing the laser and beam sources.
    psi, b_t_bar, rhi, chi : ndarray
        Arrays to be filled in while evolving the plasma column. They contain,
        respectively, the wakefield potential, the azimuthal magnetic field
        from the plasma, the plasma charge density and the plasma
        susceptibility.
    xi_fld, r_fld : ndarray
        Array containing the longitudinal and radial coordinates of the
        grid points.
    dxi, dr : float
        Longitudinal and radial step size.
    n_xi, n_r : int
        Number of grid elements in the longitudinal and radial direction.
    max_gamma : float
        Maximum gamma allowed for the plasma particles.
    p_shape : str
        Particle shape.
    plasma_pusher : str
        The plasma particle pusher.
    """

    # Calculate plasma evolution with Adams-Bashforth pusher.
    if plasma_pusher == "ab5":
        dr_arrays, dpr_arrays = pp.get_ab5_arrays()
        calculate_with_ab5(
            pp.r,
            pp.pr,
            pp.pz,
            pp.gamma,
            pp.q,
            pp.r_max_plasma,
            pp.dr_p,
            pp.parabolic_coefficient,
            *pp.get_field_arrays(),
            a2,
            nabla_a2,
            b_t_0,
            psi,
            b_t_bar,
            rho,
            chi,
            xi_fld,
            r_fld,
            dxi,
            dr,
            n_xi,
            n_r,
            max_gamma,
            p_shape,
            *dr_arrays,
            *dpr_arrays,
        )

    # Calculate plasma evolution with Runge-Kutta pusher.
    elif plasma_pusher == "rk4":
        dr_arrays, dpr_arrays = pp.get_rk4_arrays()
        calculate_with_rk4(
            pp.r,
            pp.pr,
            pp.pz,
            pp.gamma,
            pp.q,
            pp.r_max_plasma,
            pp.dr_p,
            pp.parabolic_coefficient,
            a2,
            nabla_a2,
            b_t_0,
            psi,
            b_t_bar,
            rho,
            chi,
            xi_fld,
            r_fld,
            dxi,
            dr,
            n_xi,
            n_r,
            max_gamma,
            p_shape,
            *dr_arrays,
            *dpr_arrays,
            *pp.get_rk4_field_arrays(0),
            *pp.get_rk4_field_arrays(1),
            *pp.get_rk4_field_arrays(2),
            *pp.get_rk4_field_arrays(3),
        )

    # Raise error if pusher is not recognized.
    else:
        raise ValueError("Plasma pusher '{}' not recognized.".format(plasma_pusher))


@njit_serial()
def calculate_with_ab5(
    r_pp,
    pr_pp,
    pz_pp,
    gamma_pp,
    q_pp,
    r_max_plasma,
    dr_p,
    parabolic_coefficient,
    a2_pp,
    nabla_a2_pp,
    b_t_0_pp,
    b_t_pp,
    psi_pp,
    dr_psi_pp,
    dxi_psi_pp,
    a2,
    nabla_a2,
    b_t_0,
    psi,
    b_t_bar,
    rho,
    chi,
    xi_fld,
    r_fld,
    dxi,
    dr,
    n_xi,
    n_r,
    max_gamma,
    p_shape,
    dr_1,
    dr_2,
    dr_3,
    dr_4,
    dr_5,
    dpr_1,
    dpr_2,
    dpr_3,
    dpr_4,
    dpr_5,
):
    """Calculate plasma evolution using the Adams-Bashforth pusher.

    Parameters
    ----------
    r_pp, pr_pp, pz_pp, gamma_pp, q_pp : ndarray
        Radial position, radial momentum, longitudinal momentum,
        Lorentz factor and charge of the plasma particles.
    r_max_plasma : float
        Maximum radial extent of the plasma
    dr_p : float
        Initial radial spacing between plasma particles.
    parabolic_coefficient : float
        Coefficient for the parabolic radial plasma profile.
    a2_pp, ..., dxi_psi_pp : ndarray
        Arrays where the value of the fields at the particle positions will
        be stored.
    a2, nabla_a2, b_t_0 : ndarray
        Laser and beam source fields.
    psi, b_t_bar, rho, chi : ndarray
        Arrays to be filled in during plasma evolution.
    xi_fld, r_fld : ndarray
        Grid coordinates
    dxi, dr : float
        Grid spacing
    n_xi, n_r : int
        Number of grid elements.
    max_gamma : float
        Maximum gamma of the plasma particles.
    p_shape : str
        Particle shape.
    dr_1, ..., dr_5 : ndarray
        Arrays containing the derivative of the radial position of the
        particles at the 5 slices previous to the next one.
    dpr_1, ..., dpr_5 : ndarray
        Arrays containing the derivative of the radial momentum of the
        particles at the 5 slices previous to the next one.
    """
    # Loop from the right to the left of the domain.
    for step in range(n_xi):
        slice_i = n_xi - step - 1
        xi = xi_fld[slice_i]

        # Calculate fields at the position of the particles and
        # calculate/deposit psi, b_t_bar, rho and chi at the current slice
        # of the grid.
        calculate_and_deposit_plasma_column(
            slice_i,
            xi,
            r_pp,
            pr_pp,
            pz_pp,
            gamma_pp,
            q_pp,
            r_max_plasma,
            dr_p,
            parabolic_coefficient,
            a2_pp,
            nabla_a2_pp,
            b_t_0_pp,
            b_t_pp,
            psi_pp,
            dr_psi_pp,
            dxi_psi_pp,
            a2,
            nabla_a2,
            b_t_0,
            psi,
            b_t_bar,
            rho,
            chi,
            xi_fld,
            r_fld,
            dxi,
            dr,
            n_xi,
            n_r,
            max_gamma,
            p_shape,
        )

        if slice_i > 0:
            # Evolve plasma to next xi step.
            evolve_plasma_ab5(
                dxi,
                r_pp,
                pr_pp,
                gamma_pp,
                nabla_a2_pp,
                b_t_0_pp,
                b_t_pp,
                psi_pp,
                dr_psi_pp,
                dr_1,
                dr_2,
                dr_3,
                dr_4,
                dr_5,
                dpr_1,
                dpr_2,
                dpr_3,
                dpr_4,
                dpr_5,
            )


@njit_serial()
def calculate_with_rk4(
    r_pp,
    pr_pp,
    pz_pp,
    gamma_pp,
    q_pp,
    r_max_plasma,
    dr_p,
    parabolic_coefficient,
    a2,
    nabla_a2,
    b_t_0,
    psi,
    b_t_bar,
    rho,
    chi,
    xi_fld,
    r_fld,
    dxi,
    dr,
    n_xi,
    n_r,
    max_gamma,
    p_shape,
    dr_1,
    dr_2,
    dr_3,
    dr_4,
    dpr_1,
    dpr_2,
    dpr_3,
    dpr_4,
    a2_1,
    nabla_a2_1,
    b_t_0_1,
    b_t_1,
    psi_1,
    dr_psi_1,
    dxi_psi_1,
    a2_2,
    nabla_a2_2,
    b_t_0_2,
    b_t_2,
    psi_2,
    dr_psi_2,
    dxi_psi_2,
    a2_3,
    nabla_a2_3,
    b_t_0_3,
    b_t_3,
    psi_3,
    dr_psi_3,
    dxi_psi_3,
    a2_4,
    nabla_a2_4,
    b_t_0_4,
    b_t_4,
    psi_4,
    dr_psi_4,
    dxi_psi_4,
):
    """Calculate plasma evolution using the Adams-Bashforth pusher.

    Parameters
    ----------
    r_pp, pr_pp, pz_pp, gamma_pp, q_pp : ndarray
        Radial position, radial momentum, longitudinal momentum,
        Lorentz factor and charge of the plasma particles.
    r_max_plasma : float
        Maximum radial extent of the plasma
    dr_p : float
        Initial radial spacing between plasma particles.
    parabolic_coefficient : float
        Coefficient for the parabolic radial plasma profile.
    a2, nabla_a2, b_t_0 : ndarray
        Laser and beam source fields.
    psi, b_t_bar, rho, chi : ndarray
        Arrays to be filled in during plasma evolution.
    xi_fld, r_fld : ndarray
        Grid coordinates
    dxi, dr : float
        Grid spacing
    n_xi, n_r : int
        Number of grid elements.
    max_gamma : float
        Maximum gamma of the plasma particles.
    p_shape : str
        Particle shape.
    dr_1, ..., dr_4 : ndarray
        Arrays containing the derivative of the radial position of the
        particles at the current slice and the 3 intermediate steps.
    dpr_1, ..., dpr_4 : ndarray
        Arrays containing the derivative of the radial momentum of the
        particles at the current slice and the 3 intermediate steps.
    a2_i, ..., dxi_psi_i : ndarray
        Arrays where the field values at the particle positions at substep i
        will be stored.
    """
    # Loop from the right to the left of the domain.
    for step in range(n_xi):
        slice_i = n_xi - step - 1
        xi = xi_fld[slice_i]

        # Calculate fields at the position of the particles and
        # calculate/deposit psi, b_t_bar, rho and chi at the current slice
        # of the grid.
        calculate_and_deposit_plasma_column(
            slice_i,
            xi,
            r_pp,
            pr_pp,
            pz_pp,
            gamma_pp,
            q_pp,
            r_max_plasma,
            dr_p,
            parabolic_coefficient,
            a2_1,
            nabla_a2_1,
            b_t_0_1,
            b_t_1,
            psi_1,
            dr_psi_1,
            dxi_psi_1,
            a2,
            nabla_a2,
            b_t_0,
            psi,
            b_t_bar,
            rho,
            chi,
            xi_fld,
            r_fld,
            dxi,
            dr,
            n_xi,
            n_r,
            max_gamma,
            p_shape,
        )

        if slice_i > 0:
            # Evolve plasma to next xi step.
            evolve_plasma_rk4(
                dxi,
                dr,
                xi,
                r_pp,
                pr_pp,
                gamma_pp,
                q_pp,
                r_max_plasma,
                dr_p,
                parabolic_coefficient,
                a2,
                nabla_a2,
                b_t_0,
                r_fld,
                xi_fld,
                dr_1,
                dr_2,
                dr_3,
                dr_4,
                dpr_1,
                dpr_2,
                dpr_3,
                dpr_4,
                a2_1,
                nabla_a2_1,
                b_t_0_1,
                b_t_1,
                psi_1,
                dr_psi_1,
                dxi_psi_1,
                a2_2,
                nabla_a2_2,
                b_t_0_2,
                b_t_2,
                psi_2,
                dr_psi_2,
                dxi_psi_2,
                a2_3,
                nabla_a2_3,
                b_t_0_3,
                b_t_3,
                psi_3,
                dr_psi_3,
                dxi_psi_3,
                a2_4,
                nabla_a2_4,
                b_t_0_4,
                b_t_4,
                psi_4,
                dr_psi_4,
                dxi_psi_4,
            )


@njit_serial()
def calculate_and_deposit_plasma_column(
    i,
    xi,
    r_pp,
    pr_pp,
    pz_pp,
    gamma_pp,
    q_pp,
    r_max_plasma,
    dr_p,
    parabolic_coefficient,
    a2_pp,
    nabla_a2_pp,
    b_t_0_pp,
    b_t_pp,
    psi_pp,
    dr_psi_pp,
    dxi_psi_pp,
    a2,
    nabla_a2,
    b_t_0,
    psi,
    b_t_bar,
    rho,
    chi,
    xi_fld,
    r_fld,
    dxi,
    dr,
    n_xi,
    n_r,
    max_gamma,
    p_shape,
):
    """Calculate the fields at the current position of the plasma particles
    and calculate/deposit the azimuthal magnetic field from the plasma
    (b_t_bar), the wakefield potential (psi), the plasma charge density (rho)
    and the plasma susceptibility (chi) at the current slice of the grid.

    Parameters
    ----------
    i : int
        Index of the current step.
    xi : float
        Current longitudinal position of the plasma slice.
    r_pp, pr_pp, pz_pp, gamma_pp, q_pp : ndarray
        Radial position, radial momentum, longitudinal momentum,
        Lorentz factor and charge of the plasma particles.
    r_max_plasma : float
        Maximum radial extent of the plasma.
    dr_p : float
        Initial radial spacing between plasma particles.
    parabolic_coefficient : float
        Coefficient for the parabolic radial plasma profile.
    a2_pp, ..., dxi_psi_pp : ndarray
        Arrays where the value of the fields at the particle positions will
        be stored.
    a2, nabla_a2, b_t_0 : ndarray
        Laser and beam source fields.
    psi, b_t_bar, rho, chi : ndarray
        Arrays to be filled in during plasma evolution.
    xi_fld, r_fld : ndarray
        Grid coordinates.
    dxi, dr : float
        Grid spacing.
    n_xi, n_r : int
        Number of grid elements.
    max_gamma : float
        Maximum gamma of the plasma particles.
    p_shape : str
        Particle shape.
    """
    # Gather source terms at position of plasma particles.
    gather_sources_qs_baxevanis(
        a2,
        nabla_a2,
        b_t_0,
        xi_fld[0],
        xi_fld[-1],
        r_fld[0],
        r_fld[-1],
        dxi,
        dr,
        r_pp,
        xi,
        a2_pp,
        nabla_a2_pp,
        b_t_0_pp,
    )

    # Get sorted particle indices
    idx = np.argsort(r_pp)

    # Calculate wakefield potential and derivatives at plasma particles.
    calculate_psi_and_derivatives_at_particles(
        r_pp,
        pr_pp,
        q_pp,
        idx,
        r_max_plasma,
        dr_p,
        parabolic_coefficient,
        psi_pp,
        dr_psi_pp,
        dxi_psi_pp,
    )

    # Update gamma and pz of plasma particles
    update_gamma_and_pz(gamma_pp, pz_pp, pr_pp, a2_pp, psi_pp)

    # Calculate azimuthal magnetic field from the plasma at the location of
    # the plasma particles.
    calculate_b_theta_at_particles(
        r_pp,
        pr_pp,
        q_pp,
        gamma_pp,
        psi_pp,
        dr_psi_pp,
        dxi_psi_pp,
        b_t_0_pp,
        nabla_a2_pp,
        idx,
        dr_p,
        b_t_pp,
    )

    # If particles violate the quasistatic condition, slow them down again.
    # This preserves the charge and shows better behavior than directly
    # removing them.
    idx_keep = np.where(gamma_pp >= max_gamma)
    if idx_keep[0].size > 0:
        pz_pp[idx_keep] = 0.0
        gamma_pp[idx_keep] = 1.0
        pr_pp[idx_keep] = 0.0

    # Calculate fields at specified radii for current plasma column.
    calculate_psi(r_fld, r_pp, q_pp, idx, r_max_plasma, parabolic_coefficient, psi, i)
    calculate_b_theta(
        r_fld,
        r_pp,
        pr_pp,
        q_pp,
        gamma_pp,
        psi_pp,
        dr_psi_pp,
        dxi_psi_pp,
        b_t_0_pp,
        nabla_a2_pp,
        idx,
        b_t_bar,
        i,
    )

    # Deposit rho and chi of plasma column
    w_rho = q_pp / (dr * r_pp * (1 - pz_pp / gamma_pp))
    w_chi = w_rho / gamma_pp
    deposit_plasma_particles(
        i, r_pp, w_rho, xi_fld[0], r_fld[0], n_xi, n_r, dxi, dr, rho, p_shape=p_shape
    )
    deposit_plasma_particles(
        i, r_pp, w_chi, xi_fld[0], r_fld[0], n_xi, n_r, dxi, dr, chi, p_shape=p_shape
    )


@njit_serial
def update_gamma_and_pz(gamma, pz, pr, a2, psi):
    """
    Update the gamma factor and longitudinal momentum of the plasma particles.

    Parameters
    ----------
    gamma, pz : ndarray
        Arrays containing the current gamma factor and longitudinal momentum
        of the plasma particles (will be modified here).
    pr, a2, psi : ndarray
        Arrays containing the radial momentum of the particles and the
        value of a2 and psi at the position of the particles.

    """
    for i in range(pr.shape[0]):
        gamma[i] = (1 + pr[i] ** 2 + a2[i] + (1 + psi[i]) ** 2) / (2 * (1 + psi[i]))
        pz[i] = (1 + pr[i] ** 2 + a2[i] - (1 + psi[i]) ** 2) / (2 * (1 + psi[i]))


def calculate_beam_source(bunch, n_p, n_r, n_xi, r_min, xi_min, dr, dxi, p_shape, b_t):
    """
    Return a (nz+4, nr+4) array with the azimuthal magnetic field
    from a particle distribution. This is Eq. (18) in the original paper.

    """
    # Plasma skin depth.
    s_d = ge.plasma_skin_depth(n_p / 1e6)

    # Get and normalize particle coordinate arrays.
    xi_n = bunch.xi / s_d
    x_n = bunch.x / s_d
    y_n = bunch.y / s_d

    # Calculate particle weights.
    w = bunch.q / ct.e / (2 * np.pi * dr * dxi * s_d**3 * n_p)

    # Obtain charge distribution (using cubic particle shape by default).
    q_dist = np.zeros((n_xi + 4, n_r + 4))
    deposit_3d_distribution(
        xi_n,
        x_n,
        y_n,
        w,
        xi_min,
        r_min,
        n_xi,
        n_r,
        dxi,
        dr,
        q_dist,
        p_shape=p_shape,
        use_ruyten=True,
    )

    # Remove guard cells.
    q_dist = q_dist[2:-2, 2:-2]

    # Radial position of grid points.
    r_grid_g = (0.5 + np.arange(n_r)) * dr

    # At each grid cell, calculate integral only until cell center by
    # assuming that half the charge is evenly distributed within the cell
    # (i.e., subtract half the charge)
    subs = q_dist / 2

    # At the first grid point along r, subtract an additional 1/4 of the
    # charge. This comes from assuming that the density has to be zero on axis.
    subs[:, 0] += q_dist[:, 0] / 4

    # Calculate field by integration.
    b_t[2:-2, 2:-2] += (np.cumsum(q_dist, axis=1) - subs) * dr / np.abs(r_grid_g)

    return b_t



"""Contains the 5th order Adams–Bashforth pusher for the plasma particles."""

from wake_t.utilities.numba import njit_serial


@njit_serial()
def evolve_plasma_ab2(
    dxi,
    r,
    pr,
    gamma,
    m,
    q,
    r_to_x,
    nabla_a2,
    b_theta_0,
    b_theta,
    psi,
    dr_psi,
    dr,
    dpr,
):
    """
    Evolve the r and pr coordinates of plasma particles to the next xi step
    using an Adams-Bashforth method of 2nd order.

    Parameters
    ----------
    dxi : float
        Longitudinal step.
    r, pr, gamma, r_to_x : ndarray
        Radial position, radial momentum, Lorentz factor as well an array that
        keeps track of axis crosses to convert from r to x.
    m, q : float
        Mass and charge of the plasma species.
    nabla_a2, b_theta_0, b_theta, psi, dr_psi : ndarray
        Arrays with the value of the fields at the particle positions.
    dr, dpr : ndarray
        Arrays containing the derivative of the radial position and momentum
        of the particles at the 2 slices previous to the next step.
    """

    calculate_derivatives(
        pr,
        gamma,
        m,
        q,
        b_theta_0,
        nabla_a2,
        b_theta,
        psi,
        dr_psi,
        dr[0],
        dpr[0],
    )

    # Push radial position.
    apply_ab2(r, dxi, dr)

    # Push radial momentum.
    apply_ab2(pr, dxi, dpr)

    # Shift derivatives for next step (i.e., the derivative at step i will be
    # the derivative at step i+1 in the next iteration.)
    dr[1] = dr[0]
    dpr[1] = dpr[0]

    # If a particle has crossed the axis, mirror it.
    check_axis_crossing(r, pr, dr[1], dpr[1], r_to_x)


@njit_serial(fastmath=True, error_model="numpy")
def calculate_derivatives(
    pr, gamma, m, q, b_theta_0, nabla_a2, b_theta_bar, psi, dr_psi, dr, dpr
):
    """
    Calculate the derivative of the radial position and the radial momentum
    of the plasma particles at the current slice.

    Parameters
    ----------
    pr, gamma : ndarray
        Arrays containing the radial momentum and Lorentz factor of the
        plasma particles.
    m, q : float
        Mass and charge of the plasma species.
    b_theta_0 : ndarray
        Array containing the value of the azimuthal magnetic field from
        the beam distribution at the position of each plasma particle.
    nabla_a2 : ndarray
        Array containing the value of the gradient of the laser normalized
        vector potential at the position of each plasma particle.
    b_theta_bar : ndarray
        Array containing the value of the azimuthal magnetic field from
        the plasma at the position of each plasma particle.
    psi, dr_psi : ndarray
        Arrays containing the wakefield potential and its radial derivative
        at the position of each plasma particle.
    dr, dpr : ndarray
        Arrays where the value of the derivatives of the radial position and
        radial momentum will be stored.
    """
    # Calculate derivatives of r and pr.
    q_over_m = q / m
    for i in range(pr.shape[0]):
        inv_psi_i = 1.0 / (1.0 + psi[i] * q_over_m)
        dpr[i] = (
            gamma[i] * dr_psi[i] * inv_psi_i
            - b_theta_bar[i]
            - b_theta_0[i]
            - nabla_a2[i] * 0.5 * inv_psi_i * q_over_m
        ) * q_over_m
        dr[i] = pr[i] * inv_psi_i


@njit_serial()
def apply_ab2(x, dt, dx):
    """Apply the Adams-Bashforth method of 2nd order to evolve `x`.

    Parameters
    ----------
    x : ndarray
        Array containing the variable to be advanced.
    dt : float
        Discretization step size.
    dx : ndarray
        Array containing the derivatives of `x` at the two previous steps.
    """
    for i in range(x.shape[0]):
        x[i] += dt * (1.5 * dx[0, i] - 0.5 * dx[1, i])


@njit_serial()
def check_axis_crossing(r, pr, dr, dpr, r_to_x):
    """Check for particles with r < 0 and invert them."""
    for i in range(r.shape[0]):
        if r[i] < 0.0:
            r[i] *= -1.0
            pr[i] *= -1.0
            dr[i] *= -1.0
            dpr[i] *= -1.0
            r_to_x[i] *= -1



import numpy as np
import scipy.constants as ct

from wake_t import GaussianPulse
from wake_t.fields.interpolation import interpolate_rz_field


def test_rz_interpolation_one_to_one():
    """
    This test checks that the method `interpolate_rz_field` generates
    an interpolated array that is identical to the original one when both
    arrays use the same grid.
    For this particular test, the original array is the complex envelope of
    a laser pulse.
    """
    # Grid properties.
    xi_max = 20e-6
    xi_min = -20e-6
    r_max = 400e-6
    nxi = 201
    nr = 400
    dr = r_max / nr
    dxi = (xi_max - xi_min) / (nxi - 1)

    # Time steps.
    z_tot = 1e-2
    t_tot = z_tot / ct.c
    nt = 1000
    dt = t_tot / nt

    # Laser parameters in SI units
    tau = 25e-15  # s
    w_0 = 50e-6  # m
    l_0 = 0.8e-6  # m
    z_c = 0.0  # m
    a_0 = 3
    z_foc = z_tot / 2

    # Create and initialize laser pulse.
    laser = GaussianPulse(
        z_c, l_0=l_0, w_0=w_0, a_0=a_0, tau=tau, z_foc=z_foc, polarization="circular"
    )
    laser.set_envelope_solver_params(xi_min, xi_max, r_max, nxi, nr, dt)
    laser.initialize_envelope()

    # Get laser envelope.
    a_env = laser.get_envelope()

    # Interpolate complex envelope into the same original grid.
    nxi_new = nxi
    nr_new = nr
    dr_new = r_max / nr_new
    r_min_grid = dr / 2
    xi_new = np.linspace(xi_min, xi_max, nxi_new)
    r_new = np.linspace(dr_new / 2, r_max - dr_new / 2, nr_new)
    a_env_new = np.zeros((nxi_new, nr_new), dtype=np.complex128)
    interpolate_rz_field(a_env, xi_min, r_min_grid, dxi, dr, xi_new, r_new, a_env_new)

    # Both the old and new arrays should be identical. Check that this is the
    # case.
    np.testing.assert_almost_equal(a_env_new, a_env, decimal=14)


def test_rz_interpolation_back_and_forth():
    """
    This test checks that using the `interpolate_rz_field` method to
    interpolate an array into a new grid, and then interpolate the new array
    into the original grid results in an array that is identical to the
    first one.
    This should be the case when the array values vary linearly, due to the
    linear interpolation used by the method.
    """
    # Grid properties.
    xi_max = 200e-6
    xi_min = -200e-6
    r_max = 400e-6
    nxi = 201
    nr = 400
    dr_orig = r_max / nr
    dxi_orig = (xi_max - xi_min) / (nxi - 1)

    # Create original grid.
    r_min_grid_orig = dr_orig / 2
    xi_orig = np.linspace(xi_min, xi_max, nxi)
    r_orig = np.linspace(dr_orig / 2, r_max - dr_orig / 2, nr)
    xi_grid_orig, r_grid_orig = np.meshgrid(xi_orig, r_orig, indexing="ij")

    # Create array with linear variation in both directions.
    array_orig = xi_grid_orig * 10 + r_grid_orig * 25

    # Interpolate array to a subgrid.
    nxi_new = 600
    nr_new = 200
    dr_new = r_max / nr_new
    dxi_new = (xi_max - xi_min) / (nxi_new - 1)
    r_min_grid_new = dr_new / 2
    xi_new = np.linspace(xi_min, xi_max, nxi_new)
    r_new = np.linspace(dr_new / 2, r_max - dr_new / 2, nr_new)
    array_new = np.zeros((nxi_new, nr_new))
    interpolate_rz_field(
        array_orig, xi_min, r_min_grid_orig, dxi_orig, dr_orig, xi_new, r_new, array_new
    )

    # Interpolate back to original grid.
    array_orig_interp = np.zeros((nxi, nr))
    interpolate_rz_field(
        array_new,
        xi_min,
        r_min_grid_new,
        dxi_new,
        dr_new,
        xi_orig,
        r_orig,
        array_orig_interp,
    )

    # Since the original array varies linearly, the final array should be equal
    # to the original (within precision). Check that this is true.
    np.testing.assert_almost_equal(array_orig_interp, array_orig, decimal=17)


if __name__ == "__main__":
    test_rz_interpolation_one_to_one()
    test_rz_interpolation_back_and_forth()



"""
Contains the method to compute the azimuthal magnetic field from the plasma
according to the paper by P. Baxevanis and G. Stupakov.

"""

import numpy as np

from wake_t.utilities.numba import njit_serial


@njit_serial()
def calculate_b_theta_at_particles(
    r, pr, q, gamma, psi, dr_psi, dxi_psi, b_theta_0, nabla_a2, idx, dr_p, b_theta_pp
):
    """
    Calculate the azimuthal magnetic field from the plasma at the location
    of the plasma particles using Eqs. (24), (26) and (27) from the paper
    of P. Baxevanis and G. Stupakov.

    As indicated in the original paper, the value of the fields at the
    discontinuities (at the exact radial position of the plasma particles)
    is calculated as the average between the two neighboring values.

    Parameters
    ----------
    r, pr, q, gamma : arrays
        Arrays containing, respectively, the radial position, radial momentum,
        charge and gamma (Lorentz) factor of the plasma particles.
    psi, dr_psi, dxi_psi : arrays
        Arrays with the value of the wakefield potential and its radial and
        longitudinal derivatives at the location of the plasma particles.
    b_theta_0, nabla_a2 : arrays
        Arrays with the value of the source terms. The first one being the
        azimuthal magnetic field due to the beam distribution, and the second
        the gradient of the normalized vector potential of the laser.
    idx : ndarray
        Array containing the (radially) sorted indices of the plasma particles.
    dr_p : float
        Initial spacing between plasma macroparticles. Corresponds also the
        width of the plasma sheet represented by the macroparticle.

    """
    # Calculate a_i and b_i, as well as a_0 and the sorted particle indices.
    a_i, b_i, a_0 = calculate_ai_bi_from_axis(
        r, pr, q, gamma, psi, dr_psi, dxi_psi, b_theta_0, nabla_a2, idx
    )

    # Calculate field at particles as average between neighboring values.
    n_part = r.shape[0]

    # Calculate field value at plasma particles by interpolating between two
    # neighboring values. Same as with psi and its derivaties.
    for i_sort in range(n_part):
        i = idx[i_sort]
        r_i = r[i]
        if i_sort > 0:
            r_im1 = r[idx[i_sort - 1]]
            a_im1 = a_i[idx[i_sort - 1]]
            b_im1 = b_i[idx[i_sort - 1]]
            r_left = (r_im1 + r_i) / 2
            b_theta_left = a_im1 * r_left + b_im1 / r_left
        else:
            b_theta_left = 0.0
            r_left = 0.0
        if i_sort < n_part - 1:
            r_ip1 = r[idx[i_sort + 1]]
        else:
            r_ip1 = r[i] + dr_p / 2
        r_right = (r_i + r_ip1) / 2
        b_theta_right = a_i[i] * r_right + b_i[i] / r_right

        # Do interpolation.
        b = (b_theta_right - b_theta_left) / (r_right - r_left)
        a = b_theta_left - b * r_left
        b_theta_pp[i] = a + b * r_i

        # Near the peak of a strong blowout, very large and unphysical
        # values could appear. This condition makes sure a threshold us not
        # exceeded.
        if b_theta_pp[i] > 3.0:
            b_theta_pp[i] = 3.0
        if b_theta_pp[i] < -3.0:
            b_theta_pp[i] = -3.0


@njit_serial()
def calculate_b_theta(
    r_fld, r, pr, q, gamma, psi, dr_psi, dxi_psi, b_theta_0, nabla_a2, idx, b_theta, k
):
    """
    Calculate the azimuthal magnetic field from the plasma at the radial
    locations in r_fld using Eqs. (24), (26) and (27) from the paper
    of P. Baxevanis and G. Stupakov.

    Parameters
    ----------
    r_fld : array
        Array containing the radial positions where psi should be calculated.
    r, pr, q, gamma : arrays
        Arrays containing, respectively, the radial position, radial momentum,
        charge and gamma (Lorentz) factor of the plasma particles.
    psi, dr_psi, dxi_psi : arrays
        Arrays with the value of the wakefield potential and its radial and
        longitudinal derivatives at the location of the plasma particles.
    b_theta_0, nabla_a2 : arrays
        Arrays with the value of the source terms. The first one being the
        azimuthal magnetic field due to the beam distribution, and the second
        the gradient of the normalized vector potential of the laser.
    idx : ndarray
        Array containing the (radially) sorted indices of the plasma particles.
    b_theta : ndarray
        Array where the values of the plasma azimuthal magnetic field will be
        stored.
    k : int
        Index that determines the slice of b_theta where the values will
        be filled in (the index is k+2 due to the guard cells in the array).

    """
    # Calculate a_i and b_i, as well as a_0 and the sorted particle indices.
    a_i, b_i, a_0 = calculate_ai_bi_from_axis(
        r, pr, q, gamma, psi, dr_psi, dxi_psi, b_theta_0, nabla_a2, idx
    )

    # Calculate fields at r_fld
    n_part = r.shape[0]
    n_points = r_fld.shape[0]
    b_theta_mesh = b_theta[k + 2]
    i_last = 0
    for j in range(n_points):
        r_j = r_fld[j]
        # Get index of last plasma particle with r_i < r_j, continuing from
        # last particle found in previous iteration.
        for i_sort in range(i_last, n_part):
            i_p = idx[i_sort]
            r_i = r[i_p]
            i_last = i_sort
            if r_i >= r_j:
                i_last -= 1
                break
        # Calculate fields.
        if i_last == -1:
            b_theta_mesh[2 + j] = a_0 * r_j
            i_last = 0
        else:
            i_p = idx[i_last]
            b_theta_mesh[2 + j] = a_i[i_p] * r_j + b_i[i_p] / r_j


@njit_serial()
def calculate_ai_bi_from_axis(
    r, pr, q, gamma, psi, dr_psi, dxi_psi, b_theta_0, nabla_a2, idx
):
    """
    Calculate the values of a_i and b_i which are needed to determine
    b_theta at any r position.

    For details about the input parameters see method 'calculate_b_theta'.

    The values of a_i and b_i are calculated as follows, using Eqs. (26) and
    (27) from the paper of P. Baxevanis and G. Stupakov:

        Write a_i and b_i as linear system of a_0:

            a_i = K_i * a_0_diff + T_i
            b_i = U_i * a_0_diff + P_i


        Where (im1 stands for subindex i-1):

            K_i = (1 + A_i*r_i/2) * K_im1  +  A_i/(2*r_i)     * U_im1
            U_i = (-A_i*r_i**3/2) * K_im1  +  (1 - A_i*r_i/2) * U_im1

            T_i = ( (1 + A_i*r_i/2) * T_im1  +  A_i/(2*r_i)     * P_im1  +
                    (2*Bi + Ai*Ci)/4 )
            P_i = ( (-A_i*r_i**3/2) * T_im1  +  (1 - A_i*r_i/2) * P_im1  +
                    r_i*(4*Ci - 2*Bi*r_i - Ai*Ci*r_i)/4 )

        With initial conditions:

            K_0 = 1
            U_0 = 0
            T_0 = 0
            P_0 = 0

        Then a_0 can be determined by imposing a_N = 0:

            a_N = K_N * a_0_diff + T_N = 0 <=> a_0_diff = - T_N / K_N

        If the precision of a_i and b_i becomes too low, then T_i and P_i are
        recalculated with an initial guess equal to a_i and b_i, as well as a
        new a_0_diff.

    """
    n_part = r.shape[0]

    # Preallocate arrays
    K = np.zeros(n_part)
    U = np.zeros(n_part)
    T = np.zeros(n_part)
    P = np.zeros(n_part)

    # Establish initial conditions (K_0 = 1, U_0 = 0, O_0 = 0, P_0 = 0)
    K_im1 = 1.0
    U_im1 = 0.0
    T_im1 = 0.0
    P_im1 = 0.0

    a_0 = 0.0

    for i_sort in range(n_part):
        i = idx[i_sort]
        r_i = r[i]
        q_i = q[i]
        psi_i = psi[i]

        a = 1.0 + psi_i
        b = 1.0 / (r_i * a)

        A_i = q_i * b

        l_i = 1.0 + 0.5 * A_i * r_i
        m_i = 0.5 * A_i / r_i
        n_i = -0.5 * A_i * r_i**3
        o_i = 1.0 - 0.5 * A_i * r_i

        K_i = l_i * K_im1 + m_i * U_im1
        U_i = n_i * K_im1 + o_i * U_im1

        K[i] = K_i
        U[i] = U_i

        K_im1 = K_i
        U_im1 = U_i

    i_start = 0

    while i_start < n_part:
        # Iterate over particles
        for i_sort in range(i_start, n_part):
            i = idx[i_sort]
            r_i = r[i]
            pr_i = pr[i]
            q_i = q[i]
            gamma_i = gamma[i]
            psi_i = psi[i]
            dr_psi_i = dr_psi[i]
            dxi_psi_i = dxi_psi[i]
            b_theta_0_i = b_theta_0[i]
            nabla_a2_i = nabla_a2[i]

            a = 1.0 + psi_i
            a2 = a * a
            a3 = a2 * a
            b = 1.0 / (r_i * a)
            c = 1.0 / (r_i * a2)
            pr_i2 = pr_i * pr_i

            A_i = q_i * b
            B_i = q_i * (
                -(gamma_i * dr_psi_i) * c
                + (pr_i2 * dr_psi_i) / (r_i * a3)
                + (pr_i * dxi_psi_i) * c
                + pr_i2 / (r_i * r_i * a2)
                + b_theta_0_i * b
                + nabla_a2_i * c * 0.5
            )
            C_i = q_i * (pr_i2 * c - (gamma_i / a - 1.0) / r_i)

            l_i = 1.0 + 0.5 * A_i * r_i
            m_i = 0.5 * A_i / r_i
            n_i = -0.5 * A_i * r_i**3
            o_i = 1.0 - 0.5 * A_i * r_i

            T_i = l_i * T_im1 + m_i * P_im1 + 0.5 * B_i + 0.25 * A_i * C_i
            P_i = (
                n_i * T_im1
                + o_i * P_im1
                + r_i * (C_i - 0.5 * B_i * r_i - 0.25 * A_i * C_i * r_i)
            )

            T[i] = T_i
            P[i] = P_i

            T_im1 = T_i
            P_im1 = P_i

        # Calculate a_0_diff.
        a_0_diff = -T_im1 / K_im1
        a_0 += a_0_diff

        i_stop = n_part

        # Calculate a_i (in T_i) and b_i (in P_i) as functions of a_0_diff.
        for i_sort in range(i_start, n_part):
            i = idx[i_sort]
            T_old = T[i]
            P_old = P[i]
            K_old = K[i] * a_0_diff
            U_old = U[i] * a_0_diff

            # Test if precision is lost in the sum T_old + K_old.
            # 0.5 roughly corresponds to one lost bit of precision.
            # Also pass test if this is the first number of this iteration
            # to avoid an infinite loop or if this is the last number
            # to computer as that is zero by construction
            if (
                i_sort == i_start
                or i_sort == (n_part - 1)
                or abs(T_old + K_old) >= 0.5 * abs(T_old - K_old)
                and abs(P_old + U_old) >= 0.5 * abs(P_old - U_old)
            ):
                # Calculate a_i and b_i as functions of a_0_diff.
                # Store the result in T and P
                T[i] = T_old + K_old
                P[i] = P_old + U_old
            else:
                # Stop this iteration, go to the next one
                i_stop = i_sort
                break

        if i_stop < n_part:
            # Set T_im1 and T_im1 properly for the next iteration
            T_im1 = T[idx[i_stop - 1]]
            P_im1 = P[idx[i_stop - 1]]

        # Start the next iteration where this one stopped
        i_start = i_stop

    return T, P, a_0


@njit_serial()
def calculate_ai_bi_from_edge(
    r, pr, q, gamma, psi, dr_psi, dxi_psi, b_theta_0, nabla_a2, idx
):
    """
    Calculate the values of a_i and b_i which are needed to determine
    b_theta at any r position.

    For details about the input parameters see method 'calculate_b_theta'.

    The values of a_i and b_i are calculated, using Eqs. (26) and
    (27) from the paper of P. Baxevanis and G. Stupakov. In this algorithm,
    Eq. (27) is inverted so that we calculate a_i and b_i as a function
    of a_ip1 and b_ip1. Therefore, we start the loop at the boundary and end up
    on axis. This alternative method has shown to be more robust than
    `calculate_ai_bi_from_axis` to numerical precission issues.

        Write a_i and b_i as linear system of b_N_diff:

            a_i = K_i * b_N_diff + T_i
            b_i = U_i * b_N_diff + P_i


        Where (im1 stands for subindex i-1):

            K_i = (1 - A_i*r_i/2) * K_im1  +  (-A_i/(2*r_i))  * U_im1
            U_i = A_i*r_i**3/2    * K_im1  +  (1 + A_i*r_i/2) * U_im1

            T_i = ( (1 - A_i*r_i/2) * T_im1  +  (-A_i/(2*r_i))  * P_im1  +
                    (-2*Bi + Ai*Ci)/4 )
            P_i = ( A_i*r_i**3/2    * T_im1  +  (1 + A_i*r_i/2) * P_im1  -
                    r_i*(4*Ci - 2*Bi*r_i + Ai*Ci*r_i)/4 )

        With initial conditions at i=N:

            K_Np1 = 0
            U_Np1 = 1
            T_Np1 = 0
            P_Np1 = 0

        Then b_N_diff can be determined by imposing b_0 = 0:

            b_0 = U_0 * b_N_diff + P_0 = 0
            <=> b_N_diff = - P_0 / U_0

        If the precision of a_i and b_i becomes too low, then T_i and P_i are
        recalculated with an initial guess equal to a_i and b_i, as well as a
        new b_N_diff.

    """

    n_part = r.shape[0]

    # Preallocate arrays
    K = np.zeros(n_part)
    U = np.zeros(n_part)
    T = np.zeros(n_part)
    P = np.zeros(n_part)

    # Initial conditions at i = N
    K_ip1 = 0.0
    U_ip1 = 1.0
    T_ip1 = 0.0
    P_ip1 = 0.0

    # Iterate over particles to get K_i and U_i
    for i_sort in range(n_part):
        i = idx[-1 - i_sort]
        r_i = r[i]
        q_i = q[i]
        psi_i = psi[i]

        a = 1.0 + psi_i

        l_i = 1.0 - 0.5 * q_i / a
        m_i = -0.5 * q_i / (a * r_i**2)
        n_i = 0.5 * q_i / a * r_i**2
        o_i = 1.0 + 0.5 * q_i / a

        K_i = l_i * K_ip1 + m_i * U_ip1
        U_i = n_i * K_ip1 + o_i * U_ip1

        K[i] = K_ip1
        U[i] = U_ip1

        K_ip1 = K_i
        U_ip1 = U_i

    i_start = 0

    while i_start < n_part:
        # Iterate over particles to get T_i and P_i
        for i_sort in range(i_start, n_part):
            i = idx[-1 - i_sort]
            r_i = r[i]
            pr_i = pr[i]
            q_i = q[i]
            gamma_i = gamma[i]
            psi_i = psi[i]
            dr_psi_i = dr_psi[i]
            dxi_psi_i = dxi_psi[i]
            b_theta_0_i = b_theta_0[i]
            nabla_a2_i = nabla_a2[i]

            a = 1.0 + psi_i
            a2 = a * a
            a3 = a2 * a
            b = 1.0 / (r_i * a)
            c = 1.0 / (r_i * a2)
            pr_i2 = pr_i * pr_i

            A_i = q_i * b
            B_i = q_i * (
                -(gamma_i * dr_psi_i) * c
                + (pr_i2 * dr_psi_i) / (r_i * a3)
                + (pr_i * dxi_psi_i) * c
                + pr_i2 / (r_i * r_i * a2)
                + b_theta_0_i * b
                + nabla_a2_i * c * 0.5
            )
            C_i = q_i * (pr_i2 * c - (gamma_i / a - 1.0) / r_i)

            l_i = 1.0 - 0.5 * q_i / a
            m_i = -0.5 * q_i / (a * r_i**2)
            n_i = 0.5 * q_i / a * r_i**2
            o_i = 1.0 + 0.5 * q_i / a

            T_i = l_i * T_ip1 + m_i * P_ip1 - 0.5 * B_i + 0.25 * A_i * C_i
            P_i = (
                n_i * T_ip1
                + o_i * P_ip1
                - r_i * (C_i - 0.5 * B_i * r_i + 0.25 * A_i * C_i * r_i)
            )

            T[i] = T_ip1
            P[i] = P_ip1

            T_ip1 = T_i
            P_ip1 = P_i

        # Calculate b_N_diff.
        b_N_diff = -P_ip1 / U_ip1

        i_stop = n_part

        # Calculate a_i (in T_i) and b_i (in P_i) as functions of b_N_diff.
        for i_sort in range(i_start, n_part):
            i = idx[-1 - i_sort]
            T_old = T[i]
            P_old = P[i]
            K_old = K[i] * b_N_diff
            U_old = U[i] * b_N_diff

            if (
                i_sort == i_start
                or abs(T_old + K_old) >= 0.5 * abs(T_old - K_old)
                and abs(P_old + U_old) >= 0.5 * abs(P_old - U_old)
            ):
                T[i] = T_old + K_old
                P[i] = P_old + U_old
            else:
                i_stop = i_sort
                break

        if i_stop < n_part:
            T_ip1 = T[idx[-1 - i_stop]] + K[idx[-1 - i_stop]] * b_N_diff
            P_ip1 = P[idx[-1 - i_stop]] + U[idx[-1 - i_stop]] * b_N_diff
        else:
            T_ip1 = T_ip1 + K_ip1 * b_N_diff
            P_ip1 = P_ip1 + U_ip1 * b_N_diff

        i_start = i_stop

    # Return a_i (in T_i), b_i (in P_i) and a_0 (in T_ip1)
    return T, P, T_ip1



"""
This module contains the definition of the PlasmaRamp class as well as
predefined ramp profiles.

"""

from typing import Optional, Union, Callable, Literal
from functools import partial

import numpy as np

from .plasma_stage import PlasmaStage, DtBunchType


# Define type alias for the ramp profiles.
Profile = Callable[[float, float, float, float, float], float]


def inverse_square_profile(
    z, decay_length=None, density_top=None, density_down=None, position_down=None
):
    if decay_length is None:
        decay_length = position_down / (np.sqrt(density_top / density_down) - 1)
    return density_top / np.square(1 + z / decay_length)


def exponential_profile(
    z, decay_length=None, density_top=None, density_down=None, position_down=None
):
    if decay_length is None:
        decay_length = position_down / np.log(density_top / density_down)
    return density_top * np.exp(-z / decay_length)


def gaussian_profile(
    z, decay_length=None, density_top=None, density_down=None, position_down=None
):
    if decay_length is None:
        decay_length = position_down / np.sqrt(2 * np.log(density_top / density_down))
    return density_top * np.exp(-(z**2) / (2 * decay_length**2))


ramp_profiles = {
    "inverse_square": inverse_square_profile,
    "exponential": exponential_profile,
    "gaussian": gaussian_profile,
}


class PlasmaRamp(PlasmaStage):
    """
    Class defining a plasma density ramp.

    This elements is a subclass of :class:`PlasmaStage` that exposes
    convenient attributes for defining a plasma density ramp, such as
    predefined density profiles.

    Parameters
    ----------
    length : float
        Length of the plasma ramp in m.
    profile : string or callable
        Longitudinal density profile of the ramp. Possible string values
        are ``'gaussian'``, ``'inverse square'`` and ``'exponential'``.
        A callable might also be provided as a function of the form
        ``'def func(z, decay_length, density_top, density_down,
        position_down)'`` that returns the density value at the given
        position ``z``.
    ramp_type : string
        Possible types are ``'upramp'`` and ``'downramp'``.
    wakefield_model : str
        Wakefield model to be used. Possible values are ``'blowout'``,
        ``'custom_blowout'``, ``'focusing_blowout'``, ``'cold_fluid_1d'`` and
        ``'quasistatic_2d'``.
    decay_length : float
        Optional. Characteristic decay length of the ramp. If not provided,
        it will be determined from ``plasma_dens_top``, ``plasma_dens_down``
        and ``position_down``.
    plasma_dens_top : float
        Optional. Needed only if `decay_length=None`. Plasma density at the
        beginning (end) of the downramp (upramp) in units of :math:`m^{-3}`.
    plasma_dens_down : float
        Optional. Needed only if ``decay_length=None``. Plasma density at the
        position ``position_down`` in units of :math:`m^{-3}`.
    position_down : float
        Optional. Needed only if ``decay_length=None``. Position where the
        plasma density will be equal to ``plasma_dens_down`` measured from
        the beginning (end) of the downramp (upramp). If not provided,
        the ``length`` value is assigned.
    bunch_pusher : str
        The pusher used to evolve the particle bunches in time within
        the specified fields. Possible values are ``'rk4'`` (Runge-Kutta
        method of 4th order) or ``'boris'`` (Boris method).
    dt_bunch : float
        The time step for evolving the particle bunches. If ``None``, it will
        be automatically set to :math:`dt = T/(10*2*pi)`, where T is the
        smallest expected betatron period of the bunch along the plasma stage.
        A list of values can also be provided. In this case, the list
        should have the same order as the list of bunches given to the
        ``track`` method.
    push_bunches_before_diags : bool, optional
        Whether to push the bunches before saving them to the diagnostics.
        Since the time step of the diagnostics can be different from that
        of the bunches, it could happen that the bunches appear in the
        diagnostics as they were at the last push, but not at the actual
        time of the diagnostics. Setting this parameter to ``True``
        (default) ensures that an additional push is given to all bunches
        to evolve them to the diagnostics time before saving.
        This additional push will always have a time step smaller than
        the the time step of the bunch, so it has no detrimental impact
        on the accuracy of the simulation. However, it could make
        convergence studies more difficult to interpret,
        since the number of pushes will depend on `n_diags`. Therefore,
        it is exposed as an option so that it can be disabled if needed.
    n_out : int
        Number of times along the stage in which the particle distribution
        should be returned (A list with all output bunches is returned
        after tracking).
    name : str
        Name of the plasma ramp. This is only used for displaying the
        progress bar during tracking. By default, ``'Plasma ramp'``.
    **model_params
        Keyword arguments which will be given to the wakefield model. Each
        model requires a different set of parameters. See :class:`PlasmaStage`
        documentation for more details.

    See Also
    --------
    PlasmaStage

    """

    def __init__(
        self,
        length: float,
        profile: Optional[Union[str, Profile]] = "inverse_square",
        ramp_type: Optional[str] = "upramp",
        wakefield_model: Optional[str] = "focusing_blowout",
        decay_length: Optional[float] = None,
        plasma_dens_top: Optional[float] = None,
        plasma_dens_down: Optional[float] = None,
        position_down: Optional[float] = None,
        bunch_pusher: Optional[Literal["boris", "rk4"]] = "boris",
        dt_bunch: Optional[DtBunchType] = "auto",
        push_bunches_before_diags: Optional[bool] = True,
        n_out: Optional[int] = 1,
        name: Optional[str] = "Plasma ramp",
        **model_params,
    ) -> None:
        self.ramp_type = ramp_type
        if position_down is None:
            position_down = length
        # If a function profile is not provided, generate from presets.
        if not callable(profile):
            if profile in ramp_profiles:
                profile = ramp_profiles[profile]
            else:
                raise ValueError('Ramp profile "{}" not recognized'.format(profile))
        profile = partial(
            profile,
            decay_length=decay_length,
            density_top=plasma_dens_top,
            density_down=plasma_dens_down,
            position_down=position_down,
        )
        self.profile = profile
        super().__init__(
            length=length,
            density=self.ramp_profile,
            wakefield_model=wakefield_model,
            bunch_pusher=bunch_pusher,
            dt_bunch=dt_bunch,
            push_bunches_before_diags=push_bunches_before_diags,
            n_out=n_out,
            name=name,
            **model_params,
        )

    def ramp_profile(self, z):
        """Return the density value at a certain z location."""
        # For an upramp, invert z coordinate.
        if self.ramp_type == "upramp":
            z = self.length - z
        return self.profile(z)



"""
Contains the Boris pusher

Authors: Jorge Ordóñez Carrasco, Ángel Ferran Pousa.
"""

import numpy as np
import scipy.constants as ct
from numba import prange

from wake_t.utilities.numba import njit_parallel
from wake_t.fields.gather import gather_fields


def apply_boris_pusher(bunch, fields, t, dt):
    """Evolve a particle bunch using the Boris pusher.

    Parameters
    ----------
    bunch : ParticleBunch
        The particle bunch to be evolved.
    fields : list
        List of fields within which the particle bunch will be evolved.
    t : float
        The current time.
    dt : float
        Time step by which to push the particles.
    """
    # Calculate particle species constant.
    q_over_mc = bunch.q_species / (bunch.m_species * ct.c)
    # Get the necessary arrays where the fields  will be gathered.
    ex, ey, ez, bx, by, bz = bunch.get_field_arrays()
    # Advance the particles half of one time steps.
    apply_half_position_push(
        bunch.x, bunch.y, bunch.xi, bunch.px, bunch.py, bunch.pz, dt
    )
    # Gather fields at this position.
    gather_fields(
        fields,
        bunch.x,
        bunch.y,
        bunch.xi,
        t + dt / 2,
        ex,
        ey,
        ez,
        bx,
        by,
        bz,
        bunch.name,
    )
    # Advances the momentum one time step using the gathered fields.
    push_momentum(bunch.px, bunch.py, bunch.pz, ex, ey, ez, bx, by, bz, dt, q_over_mc)
    # Completes the particles push using the updated momentum.
    apply_half_position_push(
        bunch.x, bunch.y, bunch.xi, bunch.px, bunch.py, bunch.pz, dt
    )


@njit_parallel()
def apply_half_position_push(x, y, xi, px, py, pz, dt):
    for i in prange(x.shape[0]):
        # Get particle momentum
        px_i = px[i]
        py_i = py[i]
        pz_i = pz[i]

        c_over_gamma_i = ct.c / np.sqrt(1 + (px_i**2 + py_i**2 + pz_i**2))

        # Update particle position
        x[i] += 0.5 * px_i * dt * c_over_gamma_i
        y[i] += 0.5 * py_i * dt * c_over_gamma_i
        xi[i] += 0.5 * (pz_i * c_over_gamma_i - ct.c) * dt


@njit_parallel()
def push_momentum(px, py, pz, ex, ey, ez, bx, by, bz, dt, q_over_mc):
    k = q_over_mc * dt / 2

    for i in prange(px.shape[0]):
        # Get particle momentum and fields.
        px_i = px[i]
        py_i = py[i]
        pz_i = pz[i]
        ex_i = ex[i]
        ey_i = ey[i]
        ez_i = ez[i]
        bx_i = bx[i]
        by_i = by[i]
        bz_i = bz[i]

        p_minus_x = px_i + k * ex_i
        p_minus_y = py_i + k * ey_i
        p_minus_z = pz_i + k * ez_i
        c_over_gamma_med = ct.c / np.sqrt(
            1 + (p_minus_x**2 + p_minus_y**2 + p_minus_z**2)
        )
        t_x = k * c_over_gamma_med * bx_i
        t_y = k * c_over_gamma_med * by_i
        t_z = k * c_over_gamma_med * bz_i
        cons_s = 2 / (1 + t_x**2 + t_y**2 + t_z**2)
        s_x = cons_s * t_x
        s_y = cons_s * t_y
        s_z = cons_s * t_z

        # Calculate first cross product
        p_xc1 = p_minus_x + p_minus_y * t_z - p_minus_z * t_y
        p_yc1 = p_minus_y + p_minus_z * t_x - p_minus_x * t_z
        p_zc1 = p_minus_z + p_minus_x * t_y - p_minus_y * t_x

        # Update particle momentum.
        px[i] = p_minus_x + p_yc1 * s_z - p_zc1 * s_y + k * ex_i
        py[i] = p_minus_y + p_zc1 * s_x - p_xc1 * s_z + k * ey_i
        pz[i] = p_minus_z + p_xc1 * s_y - p_yc1 * s_x + k * ez_i



"""
This module implements the methods for calculating the plasma wakefields
using the 2D r-z reduced model from P. Baxevanis and G. Stupakov.

See https://journals.aps.org/prab/abstract/10.1103/PhysRevAccelBeams.21.071301
for the full details about this model.
"""

import numpy as np
import scipy.constants as ct
import aptools.plasma_accel.general_equations as ge

from .plasma_particles import PlasmaParticles
from .utils import longitudinal_gradient, radial_gradient


def calculate_wakefields(
    laser_a2,
    r_max,
    xi_min,
    xi_max,
    n_r,
    n_xi,
    ppc,
    n_p,
    r_max_plasma=None,
    radial_density=None,
    p_shape="cubic",
    max_gamma=10.0,
    plasma_pusher="ab2",
    ion_motion=False,
    ion_mass=ct.m_p,
    free_electrons_per_ion=1,
    bunch_source_arrays=[],
    bunch_source_xi_indices=[],
    bunch_source_metadata=[],
    store_plasma_history=False,
    calculate_rho=True,
    particle_diags=[],
    fld_arrays=[],
):
    """
    Calculate the plasma wakefields generated by the given laser pulse and
    electron beam in the specified grid points.

    Parameters
    ----------
    laser_a2 : ndarray
        A (nz x nr) array containing the square of the laser envelope.
    beam_part : list
        List of numpy arrays containing the spatial coordinates and charge of
        all beam particles, i.e [x, y, xi, q].
    r_max : float
        Maximum radial position up to which plasma wakefield will be
        calculated.
    xi_min : float
        Minimum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    xi_max : float
        Maximum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    n_r : int
        Number of grid elements along r in which to calculate the wakefields.
    n_xi : int
        Number of grid elements along xi in which to calculate the wakefields.
    ppc : int (optional)
        Number of plasma particles per 1d cell along the radial direction.
    n_p : float
        On-axis plasma density in units of m^{-3}.
    r_max_plasma : float
        Maximum radial extension of the plasma column. If `None`, the plasma
        extends up to the `r_max` boundary of the simulation box.
    radial_density : callable
        Function defining the radial density profile.
    p_shape : str
        Particle shape to be used for the beam charge deposition. Possible
        values are 'linear' or 'cubic'.
    max_gamma : float
        Plasma particles whose `gamma` exceeds `max_gamma` are considered to
        violate the quasistatic condition and are put at rest (i.e.,
        `gamma=1.`, `pr=pz=0.`).
    plasma_pusher : str
        Numerical pusher for the plasma particles. Possible values are `'ab2'`.
    ion_motion : bool, optional
        Whether to allow the plasma ions to move. By default, False.
    ion_mass : float, optional
        Mass of the plasma ions. By default, the mass of a proton.
    free_electrons_per_ion : int, optional
        Number of free electrons per ion. The ion charge is adjusted
        accordingly to maintain a quasi-neutral plasma (i.e.,
        ion charge = e * free_electrons_per_ion). By default, 1.
    bunch_source_arrays : list, optional
        List containing the array from which the bunch source terms (the
        azimuthal magnetic field) will be gathered. It can be a single
        array for the whole domain, or one array per bunch when using
        adaptive grids.
    bunch_source_xi_indices : list, optional
        List containing 1d arrays that with the indices of the longitudinal
        plasma slices that can gather from them. This is needed because the
        adaptive grids might not extend the whole longitudinal domain of the
        plasma, so the plasma slices should only try to gather the source terms
        if they are available at the current slice.
    bunch_source_metadata : list, optional
        Metadata of each bunch source array.
    store_plasma_history : bool, optional
        Whether to store the plasma particle evolution. This might be needed
        for diagnostics or the use of adaptive grids. By default, False.
    calculate_rho : bool, optional
        Whether to deposit the plasma density. This might be needed for
        diagnostics. By default, False.
    particle_diags : list, optional
        List of particle quantities to save to diagnostics.
    """
    rho, rho_e, rho_i, chi, E_r, E_z, B_t, xi_fld, r_fld = fld_arrays

    # Convert to normalized units.
    s_d = ge.plasma_skin_depth(n_p * 1e-6)
    r_max = r_max / s_d
    xi_min = xi_min / s_d
    xi_max = xi_max / s_d
    dr = r_max / n_r
    dxi = (xi_max - xi_min) / (n_xi - 1)
    ppc = ppc.copy()
    ppc[:, 0] /= s_d
    r_max_plasma = r_max_plasma / s_d

    def radial_density_normalized(r):
        return radial_density(r * s_d) / n_p

    # Field node coordinates.
    r_fld = r_fld / s_d
    xi_fld = xi_fld / s_d

    # Initialize field arrays, including guard cells.
    nabla_a2 = np.zeros((n_xi + 4, n_r + 4))
    psi = np.zeros((n_xi + 4, n_r + 4))

    # Laser source.
    laser_source = laser_a2 is not None
    if laser_source:
        radial_gradient(laser_a2[2:-2, 2:-2], dr, nabla_a2[2:-2, 2:-2])

    # Calculate plasma response (including density, susceptibility, potential
    # and magnetic field)

    # Initialize plasma particles.
    pp = PlasmaParticles(
        r_max,
        r_max_plasma,
        dr,
        ppc,
        n_r,
        n_xi,
        radial_density_normalized,
        max_gamma,
        ion_motion,
        ion_mass,
        free_electrons_per_ion,
        plasma_pusher,
        p_shape,
        store_plasma_history,
        particle_diags,
    )
    pp.initialize()

    # Evolve plasma from right to left and calculate psi, b_t_bar, rho and
    # chi on a grid.
    for step in range(n_xi):
        slice_i = n_xi - step - 1

        pp.sort()

        if laser_source:
            pp.gather_laser_sources(
                laser_a2[slice_i + 2],
                nabla_a2[slice_i + 2],
                r_fld[0],
                r_fld[-1],
                dr,
            )
        pp.gather_bunch_sources(
            bunch_source_arrays,
            bunch_source_xi_indices,
            bunch_source_metadata,
            slice_i,
        )

        pp.calculate_fields()

        pp.calculate_psi_at_grid(r_fld, psi[slice_i + 2, 2:-2])
        pp.calculate_b_theta_at_grid(r_fld, B_t[slice_i + 2, 2:-2])

        if calculate_rho:
            pp.deposit_rho(
                rho[slice_i + 2],
                rho_e[slice_i + 2],
                rho_i[slice_i + 2],
                r_fld,
                n_r,
                dr,
            )
        elif "w" in particle_diags:
            pp.calculate_weights()
        if laser_source:
            pp.deposit_chi(chi[slice_i + 2], r_fld, n_r, dr)

        pp.ions_computed = True

        if store_plasma_history:
            pp.store_current_step()
        if slice_i > 0:
            pp.evolve(dxi)

    # Calculate derived fields (E_z, W_r, and E_r).
    E_0 = ge.plasma_cold_non_relativisct_wave_breaking_field(n_p * 1e-6)
    longitudinal_gradient(psi[2:-2, 2:-2], dxi, E_z[2:-2, 2:-2])
    radial_gradient(psi[2:-2, 2:-2], dr, E_r[2:-2, 2:-2])
    E_r -= B_t
    E_z *= -E_0
    E_r *= -E_0
    # B_t[:] = (b_t_bar + b_t_beam) * E_0 / ct.c
    B_t *= E_0 / ct.c
    return pp.get_history()



"""
This module contains methods for performing the interpolation of the fields
into the particle positions.

The interpolation methods are based on an adaptation of the field gathering
algorithm from FBPIC (https://github.com/fbpic/fbpic) assuming linear particle
shapes.

"""

import math
import numpy as np

from wake_t.utilities.numba import njit_serial, njit_parallel, prange


@njit_parallel()
def gather_field_cyl_linear(fld, z_min, z_max, r_min, r_max, dz, dr, x, y, z):
    """
    Interpolate a 2D field defined on an r-z grid to the particle positions
    of a 3D distribution.

    Parameters
    ----------
    fld : 2darray
        The field to be interpolated.
    z_min, z_max : float
        Position of the first and last field values along z.
    r_min,r_max : float
        Position of the first and last field values along r.
    dz, dr : float
        Grid step size along the longitudinal and radial direction.
    x, y, z : 1darray
        Coordinates of the particle distribution.

    Returns
    -------
    A 1darray with the field values at the location of each particle.

    """
    n_part = x.shape[0]

    # Preallocate output array with field values.
    fld_part = np.zeros(n_part)

    # Iterate over all particles.
    for i in prange(n_part):
        # Get particle position.
        x_i = x[i]
        y_i = y[i]
        z_i = z[i]
        r_i = math.sqrt(x_i**2 + y_i**2)

        # Gather field only if particle is within field boundaries.
        if z_i >= z_min and z_i <= z_max and r_i <= r_max:
            # Position in cell units.
            r_i_cell = (r_i - r_min) / dr + 2.0
            z_i_cell = (z_i - z_min) / dz + 2.0

            # Indices of upper and lower cells in r and z.
            ir_lower = int(math.floor(r_i_cell))
            ir_upper = ir_lower + 1
            iz_lower = int(math.floor(z_i_cell))
            iz_upper = iz_lower + 1

            # If lower r cell is below axis, assume same value as first cell.
            if ir_lower < 2:
                ir_lower = 2

            # Get field value at each bounding cell.
            fld_ll = fld[iz_lower, ir_lower]
            fld_lu = fld[iz_lower, ir_upper]
            fld_ul = fld[iz_upper, ir_lower]
            fld_uu = fld[iz_upper, ir_upper]

            # Interpolate in z.
            dz_u = iz_upper - z_i_cell
            dz_l = z_i_cell - iz_lower
            fld_z_1 = dz_u * fld_ll + dz_l * fld_ul
            fld_z_2 = dz_u * fld_lu + dz_l * fld_uu

            # Interpolate in r.
            dr_u = ir_upper - r_i_cell
            dr_l = 1 - dr_u
            fld_part[i] = dr_u * fld_z_1 + dr_l * fld_z_2
    return fld_part


@njit_parallel()
def gather_main_fields_cyl_linear(
    er,
    ez,
    bt,
    z_min,
    z_max,
    r_min,
    r_max,
    dz,
    dr,
    x,
    y,
    z,
    ex_part,
    ey_part,
    ez_part,
    bx_part,
    by_part,
    bz_part,
    r_min_gather=0.0,
):
    """
    Convenient method for interpolating at once (more efficient) the transverse
    and longitudinal wakefields.

    Parameters
    ----------
    er : 2darray
        The radial electric field.
    ez : 2darray
        The longitudinal electric field.
    bt : 2darray
        The azimuthal magnetic field.
    z_min, z_max : float
        Position of the first and last field values along z.
    r_min,r_max : float
        Position of the first and last field values along r.
    dz, dr : float
        Grid step size along the longitudinal and radial direction.
    x, y, z : 1darray
        Coordinates of the particle distribution.
    ex_part, ey_part, ez_part, bx_part, by_part, bz_part : 1darray
        Arrays where the gathered field components will be stored.
    r_min_gather : float
        The minimum radial position that particles must have in order to
        gather from the grid.

    Returns
    -------
    bool
        Whether all particles managed to gather from the grid.
    """
    n_part = x.shape[0]
    gathered = np.ones(n_part, dtype=np.bool_)

    # Iterate over all particles.
    for i in prange(n_part):
        # Get particle position.
        x_i = x[i]
        y_i = y[i]
        z_i = z[i]
        r_i = math.sqrt(x_i**2 + y_i**2)
        inv_r_i = 1.0 / r_i

        # Gather field only if particle is within field boundaries.
        if z_i >= z_min and z_i <= z_max and r_i >= r_min_gather and r_i <= r_max:
            # Position in cell units.
            r_i_cell = (r_i - r_min) / dr + 2
            z_i_cell = (z_i - z_min) / dz + 2

            # Indices of upper and lower cells in r and z.
            ir_lower = int(math.floor(r_i_cell))
            ir_upper = ir_lower + 1
            iz_lower = int(math.floor(z_i_cell))
            iz_upper = iz_lower + 1

            # If lower r cell is below axis, assume same value as first cell
            # for `ez` and sign inverse of the first cell value for `wx`. This
            # ensures that wx=0 on axis.
            wr_corr = 1
            if ir_lower < 2:
                ir_lower = 2
                wr_corr = -1

            # Get field value at each bounding cell.
            # wx_ll = wx[iz_lower, ir_lower] * wx_corr
            # wx_lu = wx[iz_lower, ir_upper]
            # wx_ul = wx[iz_upper, ir_lower] * wx_corr
            # wx_uu = wx[iz_upper, ir_upper]

            er_ll = er[iz_lower, ir_lower] * wr_corr
            er_lu = er[iz_lower, ir_upper]
            er_ul = er[iz_upper, ir_lower] * wr_corr
            er_uu = er[iz_upper, ir_upper]

            bt_ll = bt[iz_lower, ir_lower] * wr_corr
            bt_lu = bt[iz_lower, ir_upper]
            bt_ul = bt[iz_upper, ir_lower] * wr_corr
            bt_uu = bt[iz_upper, ir_upper]

            ez_ll = ez[iz_lower, ir_lower]
            ez_lu = ez[iz_lower, ir_upper]
            ez_ul = ez[iz_upper, ir_lower]
            ez_uu = ez[iz_upper, ir_upper]

            # Interpolate in z
            dz_u = iz_upper - z_i_cell
            dz_l = z_i_cell - iz_lower

            er_z_1 = dz_u * er_ll + dz_l * er_ul
            er_z_2 = dz_u * er_lu + dz_l * er_uu

            bt_z_1 = dz_u * bt_ll + dz_l * bt_ul
            bt_z_2 = dz_u * bt_lu + dz_l * bt_uu

            ez_z_1 = dz_u * ez_ll + dz_l * ez_ul
            ez_z_2 = dz_u * ez_lu + dz_l * ez_uu

            # Interpolate in r
            dr_u = ir_upper - r_i_cell
            dr_l = 1.0 - dr_u

            er_i = dr_u * er_z_1 + dr_l * er_z_2
            bt_i = dr_u * bt_z_1 + dr_l * bt_z_2

            ex_part[i] += er_i * x_i * inv_r_i
            ey_part[i] += er_i * y_i * inv_r_i
            ez_part[i] += dr_u * ez_z_1 + dr_l * ez_z_2
            bx_part[i] += -bt_i * y_i * inv_r_i
            by_part[i] += bt_i * x_i * inv_r_i
        else:
            gathered[i] = False
    return not np.any(gathered)


@njit_serial()
def gather_sources_qs_baxevanis(
    fld_1,
    fld_2,
    fld_3,
    z_min,
    z_max,
    r_min,
    r_max,
    dz,
    dr,
    r,
    z,
    fld_1_pp,
    fld_2_pp,
    fld_3_pp,
):
    """
    Convenient method for gathering at once the three source fields needed
    by the Baxevanis wakefield model (a2 and nabla_a from the laser, and
    b_theta_0 from the beam) into the plasma particles. This method is also
    catered to cylindrical geometry and assumes all plasma particles have the
    same longitudinal position (which is the case in a quasistatic model,
    where there is only a single column of particles).

    Parameters
    ----------
    fld_1, fld_2, fld_3 : ndarray
        The three source fields, corresponding respectively to a2, nabla_a2
        and b_theta_0. Each of them is a (nr+4, nr+4) array, including 2 guard
        cells in each boundary.
    z_min, z_max : float
        Position of the first and last field values along z.
    r_min, r_max : float
        Position of the first and last field values along r.
    dz, dr : float
        Grid step size along the longitudinal and radial direction.
    r : 1darray
        Transverse position of the plasma particles.
    z : int
        Longitudinal position of the column of plasma particles.

    Returns
    --------
    A tuple with three 1darray containing the gathered field values at the
    position of each particle.

    """

    # Iterate over all particles.
    for i in range(r.shape[0]):
        # Get particle position.
        z_i = z
        r_i = r[i]

        # Gather field only if particle is within field boundaries.
        if z_i >= z_min and z_i <= z_max and r_i <= r_max:
            # Position in cell units.
            r_i_cell = (r_i - r_min) / dr + 2
            z_i_cell = (z_i - z_min) / dz + 2

            # Indices of upper and lower cells in r and z.
            ir_lower = int(math.floor(r_i_cell))
            ir_upper = ir_lower + 1
            iz_lower = int(math.floor(z_i_cell))
            iz_upper = iz_lower + 1

            # If lower r cell is below axis, assume same value as first cell.
            # For `nabla_a2` and `b_theta_0`, invert the sign to ensure they
            # are `0` on axis.
            sign = 1
            if ir_lower < 2:
                ir_lower = 2
                sign = -1

            # Get field value at each bounding cell.
            fld_1_ll = fld_1[iz_lower, ir_lower]
            fld_1_lu = fld_1[iz_lower, ir_upper]
            fld_1_ul = fld_1[iz_upper, ir_lower]
            fld_1_uu = fld_1[iz_upper, ir_upper]
            fld_2_ll = fld_2[iz_lower, ir_lower] * sign
            fld_2_lu = fld_2[iz_lower, ir_upper]
            fld_2_ul = fld_2[iz_upper, ir_lower] * sign
            fld_2_uu = fld_2[iz_upper, ir_upper]
            fld_3_ll = fld_3[iz_lower, ir_lower] * sign
            fld_3_lu = fld_3[iz_lower, ir_upper]
            fld_3_ul = fld_3[iz_upper, ir_lower] * sign
            fld_3_uu = fld_3[iz_upper, ir_upper]

            # Interpolate in z
            dz_u = iz_upper - z_i_cell
            dz_l = z_i_cell - iz_lower
            fld_1_z_1 = dz_u * fld_1_ll + dz_l * fld_1_ul
            fld_1_z_2 = dz_u * fld_1_lu + dz_l * fld_1_uu
            fld_2_z_1 = dz_u * fld_2_ll + dz_l * fld_2_ul
            fld_2_z_2 = dz_u * fld_2_lu + dz_l * fld_2_uu
            fld_3_z_1 = dz_u * fld_3_ll + dz_l * fld_3_ul
            fld_3_z_2 = dz_u * fld_3_lu + dz_l * fld_3_uu

            # Interpolate in r
            dr_u = ir_upper - r_i_cell
            dr_l = 1 - dr_u
            fld_1_pp[i] = dr_u * fld_1_z_1 + dr_l * fld_1_z_2
            fld_2_pp[i] = dr_u * fld_2_z_1 + dr_l * fld_2_z_2
            fld_3_pp[i] = dr_u * fld_3_z_1 + dr_l * fld_3_z_2



"""Defines a progress bar to be used by the Tracker."""

import sys
import warnings

from tqdm import tqdm


# Avoid showing clamping warnings from the progress bar.
warnings.filterwarnings(
    "ignore",
    ".*clamping.*",
)


def get_progress_bar(description, total_length, disable):
    """Get progress bar for the tracker.

    Parameters
    ----------
    description : str
        Description to be appended to start of the progress bar.
    total_length : float
        Total length in metres of the stage to be tracked.
    disable : bool
        Whether to disable (not show) the progress bar.

    Returns
    -------
    A tqdm progress bar.
    """
    l_bar = "{desc}: {percentage:3.0f}%|"
    r_bar = "| {n:.6f}/{total:.6f} {unit} [{elapsed}]"
    progress_bar = tqdm(
        desc=description,
        total=total_length,
        unit="m",
        bar_format=l_bar + "{bar}" + r_bar,
        file=sys.stdout,
        disable=disable,
    )
    return progress_bar



"""
This module contains the methods for depositing particle weights on a 1D grid.

"""

import math

from wake_t.utilities.numba import njit_serial


@njit_serial()
def deposit_plasma_particles(r, w, r_min, nr, dr, deposition_array, p_shape="cubic"):
    """
    Deposit the the weight of a 1D slice of plasma particles into a 2D
    r-z grid.

    Parameters
    ----------
    r : array
        Arrays containing the radial coordinates of the
        particles.
    w : array
        Weight of the particles (any quantity) which will be deposited into
        the grid.
    r_min : float
        Position of the first field value along r.
    nr : int
        Number of grid cells (excluding guard cells) along the radial
        direction.
    dr : float
        Grid step size along the radial direction.
    deposition_array : array
        The 2D array of size nr+4 (including two guard cells at each
        boundary) into which the weight will be deposited.
    p_shape : str
        Particle shape to be used. Possible values are 'linear' or 'cubic'.

    """
    if p_shape == "linear":
        return deposit_plasma_particles_linear(r, w, r_min, nr, dr, deposition_array)
    elif p_shape == "cubic":
        return deposit_plasma_particles_cubic(r, w, r_min, nr, dr, deposition_array)


@njit_serial(fastmath=True, error_model="numpy")
def deposit_plasma_particles_linear(r, q, r_min, nr, dr, deposition_array):
    """Calculate charge distribution assuming linear particle shape."""

    r_max = nr * dr

    # Loop over particles.
    for i in range(r.shape[0]):
        # Get particle components.
        r_i = r[i]
        w_i = q[i]

        # Deposit only if particle is within field boundaries.
        if r_i <= r_max:
            # Positions of the particles in cell units.
            r_cell = (r_i - r_min) / dr

            # Indices of lowest cell in which the particle will deposit charge.
            ir_cell = int(math.ceil(r_cell)) + 1

            # u_r: particle position wrt left neighbor gridpoint in r.
            u_r = r_cell + 2 - ir_cell

            # Precalculate quantities.
            rsl_0 = 1.0 - u_r
            rsl_1 = u_r

            # Add contribution of particle to density array.
            deposition_array[ir_cell + 0] += rsl_0 * w_i
            deposition_array[ir_cell + 1] += rsl_1 * w_i

        # Apply correction on axis (ensures uniform density in a uniform
        # plasma)
        deposition_array[2] -= deposition_array[1]
        deposition_array[1] = 0.0

    for i in range(nr):
        deposition_array[i + 2] /= (r_min + i * dr) * dr


@njit_serial(fastmath=True, error_model="numpy")
def deposit_plasma_particles_cubic(r, q, r_min, nr, dr, deposition_array):
    """Calculate charge distribution assuming cubic particle shape."""

    r_max = nr * dr

    # Loop over particles.
    for i in range(r.shape[0]):
        # Get particle components.
        r_i = r[i]
        w_i = q[i]

        # Deposit only if particle is within field boundaries.
        if r_i <= r_max:
            # Positions of the particle in cell units.
            r_cell = (r_i - r_min) / dr

            # Indices of lowest cell in which the particle will deposit charge.
            ir_cell = int(math.ceil(r_cell))

            # Particle position wrt left neighbor gridpoint.
            u_r = r_cell - ir_cell + 1

            # Precalculate quantities for shape coefficients.
            inv_6 = 1.0 / 6.0
            v_r = 1.0 - u_r

            # Cubic particle shape coefficients in z and r.
            rsc_0 = inv_6 * v_r**3
            rsc_1 = inv_6 * (3.0 * u_r**3 - 6.0 * u_r**2 + 4.0)
            rsc_2 = inv_6 * (3.0 * v_r**3 - 6.0 * v_r**2 + 4.0)
            rsc_3 = inv_6 * u_r**3

            # Add contribution of particle to density array.
            deposition_array[ir_cell + 0] += rsc_0 * w_i
            deposition_array[ir_cell + 1] += rsc_1 * w_i
            deposition_array[ir_cell + 2] += rsc_2 * w_i
            deposition_array[ir_cell + 3] += rsc_3 * w_i

        # Apply correction on axis (ensures uniform density in a uniform
        # plasma)
        deposition_array[2] -= deposition_array[1]
        deposition_array[3] -= deposition_array[0]
        deposition_array[0] = 0.0
        deposition_array[1] = 0.0

    for i in range(nr):
        deposition_array[i + 2] /= (r_min + i * dr) * dr



"""Contains various utility functions for the gridless solver."""

import math

import numpy as np

from wake_t.utilities.numba import njit_serial


@njit_serial(fastmath=True)
def log(input, output):
    """Calculate log of `input` and store it in `output`"""
    for i in range(input.shape[0]):
        output[i] = math.log(input[i])


@njit_serial(error_model="numpy")
def calculate_chi(q, w, pz, gamma, chi):
    """Calculate the contribution of each particle to `chi`."""
    for i in range(w.shape[0]):
        w_i = w[i]
        pz_i = pz[i]
        inv_gamma_i = 1.0 / gamma[i]
        chi[i] = q * w_i / (1.0 - pz_i * inv_gamma_i) * inv_gamma_i


@njit_serial(error_model="numpy")
def calculate_rho(q, w, pz, gamma, rho):
    """Calculate the contribution of each particle to `rho`."""
    for i in range(w.shape[0]):
        w_i = w[i]
        pz_i = pz[i]
        inv_gamma_i = 1.0 / gamma[i]
        rho[i] = q * w_i / (1.0 - pz_i * inv_gamma_i)


@njit_serial()
def longitudinal_gradient(f, dz, dz_f):
    """Calculate the longitudinal gradient of a 2D array.

    This method is equivalent to using `np.gradient` with
    `edge_order=2`, but is several times faster as it is compiled with numba
    and is more specialized.

    Parameters
    ----------
    f : ndarray
        The array from which to calculate the gradient.
    dz : float
        Longitudinal step size.
    dz_f : ndarray
        Array where the longitudinal gradient will be stored.
    """
    nz, nr = f.shape
    inv_dz = 1.0 / dz
    inv_h = 0.5 * inv_dz
    a = -1.5 * inv_dz
    b = 2.0 * inv_dz
    c = -0.5 * inv_dz
    for j in range(nr):
        f_right = f[nz - 1, j]
        for i in range(1, nz - 1):
            f_left = f[i - 1, j]
            f_right = f[i + 1, j]
            dz_f[i, j] = (f_right - f_left) * inv_h
        dz_f[0, j] = a * f[0, j] + b * f[1, j] + c * f[2, j]
        dz_f[-1, j] = -a * f[-1, j] - b * f[-2, j] - c * f[-3, j]


@njit_serial()
def radial_gradient(f, dr, dr_f):
    """Calculate the radial gradient of a 2D array.

    This method is equivalent to using `np.gradient` with
    `edge_order=2`, but is several times faster as it is compiled with numba
    and is more specialized. It takes advantage of the axial symmetry to
    calculate the derivative on axis.

    Parameters
    ----------
    f : ndarray
        The array from which to calculate the gradient.
    dr : float
        Radial step size.
    dr_f : ndarray
        Array where the radial gradient will be stored.
    """
    nz, nr = f.shape
    inv_dr = 1.0 / dr
    inv_h = 0.5 * inv_dr
    a = -1.5 * inv_dr
    b = 2.0 * inv_dr
    c = -0.5 * inv_dr
    for i in range(nz):
        for j in range(1, nr - 1):
            f_left = f[i, j - 1]
            f_right = f[i, j + 1]
            dr_f[i, j] = (f_right - f_left) * inv_h
        dr_f[i, 0] = (f[i, 1] - f[i, 0]) * inv_h
        dr_f[i, -1] = -a * f[i, -1] - b * f[i, -2] - c * f[i, -3]


@njit_serial()
def calculate_laser_a2(a_complex, a2):
    """Calculate the square of the laser complex envelope amplitude.

    Parameters
    ----------
    a_complex : ndarray
        Array of size (nz, nr) containing the complex envelope of the laser.
    a2 : ndarray
        Array of size (nz+4, nr+4) where the result will be stored.
    """
    nz, nr = a_complex.shape
    a_real = a_complex.real
    a_imag = a_complex.imag
    for i in range(nz):
        for j in range(nr):
            ar = a_real[i, j]
            ai = a_imag[i, j]
            a2[2 + i, 2 + j] = ar * ar + ai * ai


@njit_serial(error_model="numpy")
def update_gamma_and_pz(gamma, pz, pr, a2, psi, q, m):
    """
    Update the gamma factor and longitudinal momentum of the plasma particles.

    Parameters
    ----------
    gamma, pz : ndarray
        Arrays containing the current gamma factor and longitudinal momentum
        of the plasma particles (will be modified here).
    pr, a2, psi : ndarray
        Arrays containing the radial momentum of the particles and the
        value of a2 and psi at the position of the particles.
    q, m : float
        Charge and mass of the plasma species.

    """
    q_over_m = q / m
    for i in range(pr.shape[0]):
        psi_i = psi[i] * q_over_m
        pz_i = (1 + pr[i] ** 2 + q_over_m**2 * a2[i] - (1 + psi_i) ** 2) / (
            2 * (1 + psi_i)
        )
        pz[i] = pz_i
        gamma[i] = 1.0 + pz_i + psi_i


@njit_serial()
def check_gamma(gamma, pz, pr, max_gamma):
    """Check that the gamma of particles does not exceed `max_gamma`"""
    for i in range(gamma.shape[0]):
        if gamma[i] > max_gamma:
            gamma[i] = 1.0
            pz[i] = 0.0
            pr[i] = 0.0


@njit_serial()
def sort_particle_arrays(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, indices):
    """Sort all the particle arrays with a given sorting order.

    Implementing it like this looks very ugly, but it is much faster than
    repeating `array = array[indices]` for each array. It is also much faster
    than implementing a `sort_array` function that is called on each array.
    This is probably because of the overhead from calling numba functions.
    """
    a1_orig = np.copy(a1)
    a2_orig = np.copy(a2)
    a3_orig = np.copy(a3)
    a4_orig = np.copy(a4)
    a5_orig = np.copy(a5)
    a6_orig = np.copy(a6)
    a7_orig = np.copy(a7)
    a8_orig = np.copy(a8)
    a9_orig = np.copy(a9)
    a10_orig = np.copy(a10)
    a11_orig = np.copy(a11)
    n_part = indices.shape[0]
    for i in range(n_part):
        i_sort = indices[i]
        if i != i_sort:
            a1[i] = a1_orig[i_sort]
            a2[i] = a2_orig[i_sort]
            a3[i] = a3_orig[i_sort]
            a4[i] = a4_orig[i_sort]
            a5[i] = a5_orig[i_sort]
            a6[i] = a6_orig[i_sort]
            a7[i] = a7_orig[i_sort]
            a8[i] = a8_orig[i_sort]
            a9[i] = a9_orig[i_sort]
            a10[:, i] = a10_orig[:, i_sort]
            a11[:, i] = a11_orig[:, i_sort]



import numpy as np


def first_order_matrix(z, L, theta, k1, gamma_ref):
    """
    Calculate the first order matrix for the transfer map.
    This function is an adaptation of the one found in the particle tracking
    code Ocelot (see https://github.com/ocelot-collab/ocelot) written by
    S. Tomin.

    Parameters
    ----------
    z : float
        Longitudinal position in which to calculate the transfer matrix
    L : float
        Total length of the beamline element
    theta : float
        Bending angle of the beamline element
    k1 : float
        Quadrupole gradient of the beamline element in units of 1/m^2. A
        positive value implies focusing on the 'x' plane, while a negative
        gradient corresponds to focusing on 'y'
    k2 : float
        Sextupole gradient of the beamline element in units of 1/m^3. A
        positive value implies focusing on the 'x' plane, while a negative
        gradient corresponds to focusing on 'y'
    gamma_ref : float
        Reference energy with respect to which the particle momentum dp is
        calculated.

    """
    gamma = gamma_ref
    hx = theta / L
    kx2 = k1 + hx * hx
    ky2 = -k1
    kx = np.sqrt(kx2 + 0.0j)
    ky = np.sqrt(ky2 + 0.0j)
    cx = np.cos(z * kx).real
    cy = np.cos(z * ky).real
    sy = (np.sin(ky * z) / ky).real if ky != 0 else z
    igamma2 = 0.0
    if gamma != 0:
        igamma2 = 1.0 / (gamma * gamma)
    beta = np.sqrt(1.0 - igamma2)
    if kx != 0:
        sx = (np.sin(kx * z) / kx).real
        dx = hx / kx2 * (1.0 - cx)
        r56 = hx * hx * (z - sx) / kx2 / beta**2
    else:
        sx = z
        dx = z * z * hx / 2.0
        r56 = hx * hx * z**3 / 6.0 / beta**2
    r56 -= z / (beta * beta) * igamma2
    u_matrix = np.array(
        [
            [cx, sx, 0.0, 0.0, 0.0, dx / beta],
            [-kx2 * sx, cx, 0.0, 0.0, 0.0, sx * hx / beta],
            [0.0, 0.0, cy, sy, 0.0, 0.0],
            [0.0, 0.0, -ky2 * sy, cy, 0.0, 0.0],
            [hx * sx / beta, dx / beta, 0.0, 0.0, 1.0, r56],
            [0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
        ]
    )
    return u_matrix


def second_order_matrix(z, L, theta, k1, k2, gamma_ref):
    """
    Calculate the second order matrix for the transfer map.
    This function is an adaptation of the one found in the particle tracking
    code Ocelot (see https://github.com/ocelot-collab/ocelot) written by
    S. Tomin.

    Parameters
    ----------
    beam_matrix : array
        6 x N matrix, where N is the number of particles, containing the
        phase-space information of the bunch as (x, x', y, y', xi, dp) in
        units of (m, rad, m, rad, m, -). dp is defined as
        dp = (g-g_ref)/g_ref, while x' = px/p_kin and y' = py/p_kin, where
        p_kin is the reference kinetic momentum.
    z : float
        Longitudinal position in which to obtain the bunch distribution
    L : float
        Total length of the beamline element
    theta : float
        Bending angle of the beamline element
    k1 : float
        Quadrupole gradient of the beamline element in units of 1/m^2. A
        positive value implies focusing on the 'x' plane, while a negative
        gradient corresponds to focusing on 'y'
    k2 : float
        Sextupole gradient of the beamline element in units of 1/m^3. A
        positive value implies focusing on the 'x' plane, while a negative
        gradient corresponds to focusing on 'y'
    gamma_ref : float
        Reference energy with respect to which the particle momentum dp is
        calculated.

    """
    igamma2 = 0.0
    if gamma_ref != 0:
        gamma = gamma_ref
        gamma2 = gamma * gamma
        igamma2 = 1.0 / gamma2

    beta = np.sqrt(1.0 - igamma2)
    beta2 = beta * beta
    beta3 = beta2 * beta
    h = theta / L
    L = z  # Easy fix for the calculation below, where L should actually be z.
    h2 = h * h
    h3 = h2 * h
    kx2 = k1 + h * h
    ky2 = -k1
    kx4 = kx2 * kx2
    kx = np.sqrt(kx2 + 0.0j)
    ky = np.sqrt(-k1 + 0.0j)
    cx = np.cos(kx * L).real
    sx = (np.sin(kx * L) / kx).real if kx != 0 else L
    cy = np.cos(ky * L).real

    sy = (np.sin(ky * L) / ky).real if ky != 0 else L

    sx2 = sx * sx
    sy2 = sy * sy
    L2 = L * L
    L3 = L2 * L
    L4 = L3 * L
    L5 = L4 * L
    dx = (1.0 - cx) / kx2 if kx != 0.0 else L * L / 2.0
    dx2 = dx * dx
    K1 = k1
    K2 = k2

    d2y = 0.5 * sy * sy
    s2y = sy * cy
    c2y = np.cos(2 * ky * L).real
    fx = (L - sx) / kx2 if kx2 != 0 else L3 / 6.0
    f2y = (L - s2y) / ky2 if ky2 != 0 else L3 / 6
    J1 = (L - sx) / kx2 if kx2 != 0 else L3 / 6.0
    J2 = (3.0 * L - 4.0 * sx + sx * cx) / (2 * kx4) if kx != 0 else L**5 / 20.0
    J3 = (
        (
            (15 * L - 22.5 * sx + 9 * sx * cx - 1.5 * sx * cx * cx + kx2 * sx2 * sx)
            / (6 * kx4 * kx2)
        )
        if kx != 0
        else L**7 / 56.0
    )
    # J3 = (15 * L - 22 * sx + 9 * sx * cx - 2 * sx * cx * cx ) / (
    #        6 * kx4 * kx2) if kx != 0 else L ** 7 / 56.
    J_denom = kx2 - 4 * ky2
    Jc = (c2y - cx) / J_denom if J_denom != 0 else 0.5 * L2
    Js = (cy * sy - sx) / J_denom if J_denom != 0 else L3 / 6
    Jd = (d2y - dx) / J_denom if J_denom != 0 else L4 / 24
    Jf = (f2y - fx) / J_denom if J_denom != 0 else L5 / 120

    khk = K2 + 2 * h * K1
    T = np.zeros((6, 6, 6))

    T111 = -1 / 6 * khk * (sx2 + dx) - 0.5 * h * kx2 * sx2
    T112 = -1 / 6 * khk * sx * dx + 0.5 * h * sx * cx
    T122 = -1 / 6 * khk * dx2 + 0.5 * h * dx * cx
    T116 = (
        -h / 12 / beta * khk * (3 * sx * J1 - dx2)
        + 0.5 * h2 / beta * sx2
        + 0.25 / beta * K1 * L * sx
    )
    T126 = (
        -h / 12 / beta * khk * (sx * dx2 - 2 * cx * J2)
        + 0.25 * h2 / beta * (sx * dx + cx * J1)
        - 0.25 / beta * (sx + L * cx)
    )
    T166 = (
        -h2 / 6 / beta2 * khk * (dx2 * dx - 2 * sx * J2)
        + 0.5 * h3 / beta2 * sx * J1
        - 0.5 * h / beta2 * L * sx
        - 0.5 * h / (beta2) * igamma2 * dx
    )
    T133 = K1 * K2 * Jd + 0.5 * (K2 + h * K1) * dx
    T134 = 0.5 * K2 * Js
    T144 = K2 * Jd - 0.5 * h * dx

    T211 = -1 / 6 * khk * sx * (1 + 2 * cx)
    T212 = -1 / 6 * khk * dx * (1 + 2 * cx)
    T222 = -1 / 3 * khk * sx * dx - 0.5 * h * sx
    T216 = -h / 12 / beta * khk * (3 * cx * J1 + sx * dx) - 0.25 / beta * K1 * (
        sx - L * cx
    )
    T226 = -h / 12 / beta * khk * (3 * sx * J1 + dx2) + 0.25 / beta * K1 * L * sx
    T266 = (
        -h2 / 6 / beta2 * khk * (sx * dx2 - 2 * cx * J2)
        - 0.5 * h / beta2 * K1 * (cx * J1 - sx * dx)
        - 0.5 * h / beta2 * igamma2 * sx
    )
    T233 = K1 * K2 * Js + 0.5 * (K2 + h * K1) * sx
    T234 = 0.5 * K2 * Jc
    T244 = K2 * Js - 0.5 * h * sx

    T313 = 0.5 * K2 * (cy * Jc - 2 * K1 * sy * Js) + 0.5 * h * K1 * sx * sy
    T314 = 0.5 * K2 * (sy * Jc - 2 * cy * Js) + 0.5 * h * sx * cy
    T323 = 0.5 * K2 * (cy * Js - 2 * K1 * sy * Jd) + 0.5 * h * K1 * dx * sy
    T324 = 0.5 * K2 * (sy * Js - 2 * cy * Jd) + 0.5 * h * dx * cy
    T336 = (
        0.5 * h / beta * K2 * (cy * Jd - 2 * K1 * sy * Jf)
        + 0.5 * h2 / beta * K1 * J1 * sy
        - 0.25 / beta * K1 * L * sy
    )
    T346 = (
        0.5 * h / beta * K2 * (sy * Jd - 2 * cy * Jf)
        + 0.5 * h2 / beta * J1 * cy
        - 0.25 / beta * (sy + L * cy)
    )

    T413 = 0.5 * K1 * K2 * (2 * cy * Js - sy * Jc) + 0.5 * (K2 + h * K1) * sx * cy
    T414 = 0.5 * K2 * (2 * K1 * sy * Js - cy * Jc) + 0.5 * (K2 + h * K1) * sx * sy
    T423 = 0.5 * K1 * K2 * (2 * cy * Jd - sy * Js) + 0.5 * (K2 + h * K1) * dx * cy
    T424 = 0.5 * K2 * (2 * K1 * sy * Jd - cy * Js) + 0.5 * (K2 + h * K1) * dx * sy
    T436 = (
        0.5 * h / beta * K1 * K2 * (2 * cy * Jf - sy * Jd)
        + 0.5 * h / beta * (K2 + h * K1) * J1 * cy
        + 0.25 / beta * K1 * (sy - L * cy)
    )
    T446 = (
        0.5 * h / beta * K2 * (2 * K1 * sy * Jf - cy * Jd)
        + 0.5 * h / beta * (K2 + h * K1) * J1 * sy
        - 0.25 / beta * K1 * L * sy
    )

    T511 = h / 12 / beta * khk * (sx * dx + 3 * J1) - 0.25 / beta * K1 * (L - sx * cx)
    T512 = h / 12 / beta * khk * dx2 + 0.25 / beta * K1 * sx2
    T522 = h / 6 / beta * khk * J2 - 0.5 / beta * sx - 0.25 / beta * K1 * (J1 - sx * dx)

    T516 = (
        h2 / 12 / beta2 * khk * (3 * dx * J1 - 4 * J2)
        + 0.25 * h / beta2 * K1 * J1 * (1 + cx)
        + 0.5 * h / beta2 * igamma2 * sx
    )

    T526 = (
        h2 / 12 / beta2 * khk * (dx * dx2 - 2 * sx * J2)
        + 0.25 * h / beta2 * K1 * sx * J1
        + 0.5 * h / beta2 * igamma2 * dx
    )
    T566 = (
        h3 / 6 / beta3 * khk * (3 * J3 - 2 * dx * J2)
        + h2 / 6 / beta3 * K1 * (sx * dx2 - J2 * (1 + 2 * cx))
        + 1.5 / beta3 * igamma2 * (h2 * J1 - L)
    )
    T533 = (
        -h / beta * K1 * K2 * Jf
        - 0.5 * h / beta * (K2 + h * K1) * J1
        + 0.25 / beta * K1 * (L - cy * sy)
    )
    T534 = -0.5 * h / beta * K2 * Jd - 0.25 / beta * K1 * sy2
    T544 = -h / beta * K2 * Jf + 0.5 * h2 / beta * J1 - 0.25 / beta * (L + cy * sy)

    T[0, 0, 0] = T111
    T[0, 0, 1] = T112 * 2
    T[0, 1, 1] = T122
    T[0, 0, 5] = T116 * 2
    T[0, 1, 5] = T126 * 2
    T[0, 5, 5] = T166
    T[0, 2, 2] = T133
    T[0, 2, 3] = T134 * 2
    T[0, 3, 3] = T144

    T[1, 0, 0] = T211
    T[1, 0, 1] = T212 * 2
    T[1, 1, 1] = T222
    T[1, 0, 5] = T216 * 2
    T[1, 1, 5] = T226 * 2
    T[1, 5, 5] = T266
    T[1, 2, 2] = T233
    T[1, 2, 3] = T234 * 2
    T[1, 3, 3] = T244

    T[2, 0, 2] = T313 * 2
    T[2, 0, 3] = T314 * 2
    T[2, 1, 2] = T323 * 2
    T[2, 1, 3] = T324 * 2
    T[2, 2, 5] = T336 * 2
    T[2, 3, 5] = T346 * 2

    T[3, 0, 2] = T413 * 2
    T[3, 0, 3] = T414 * 2
    T[3, 1, 2] = T423 * 2
    T[3, 1, 3] = T424 * 2
    T[3, 2, 5] = T436 * 2
    T[3, 3, 5] = T446 * 2

    T[4, 0, 0] = -T511
    T[4, 0, 1] = -T512 * 2
    T[4, 1, 1] = -T522
    T[4, 0, 5] = -T516 * 2
    T[4, 1, 5] = -T526 * 2
    T[4, 5, 5] = -T566
    T[4, 2, 2] = -T533
    T[4, 2, 3] = -T534 * 2
    T[4, 3, 3] = -T544
    return T



"""Contains other utilities"""

import sys

import numpy as np


def print_progress_bar(pre_string, step, total_steps):
    n_dash = int(round(step / total_steps * 20))
    n_space = 20 - n_dash
    status = pre_string + "[" + "-" * n_dash + " " * n_space + "] "
    if step < total_steps:
        status += "\r"
    sys.stdout.write(status)
    sys.stdout.flush()


def generate_field_diag_dictionary(
    fld_names,
    fld_comps,
    fld_attrs,
    fld_arrays,
    fld_comp_pos,
    grid_labels,
    grid_spacing,
    grid_global_offset,
    fld_solver,
    fld_solver_params,
    fld_boundary,
    fld_boundary_params,
    part_boundary,
    part_boundary_params,
    current_smoothing,
    charge_correction,
):
    """
    Generates a dictionary which can be used by the openPMD diagnostics to
    write the field data.

    """
    diag_data = {}
    diag_data["fields"] = fld_names
    fld_zip = zip(fld_names, fld_comps, fld_attrs, fld_arrays, fld_comp_pos)
    for fld, comps, attrs, arrays, pos in fld_zip:
        diag_data[fld] = {}
        if comps is not None:
            diag_data[fld]["comps"] = {}
            for comp, arr in zip(comps, arrays):
                diag_data[fld]["comps"][comp] = {}
                diag_data[fld]["comps"][comp]["array"] = arr
                diag_data[fld]["comps"][comp]["position"] = pos
        else:
            diag_data[fld]["array"] = arrays[0]
            diag_data[fld]["position"] = pos
        diag_data[fld]["grid"] = {}
        diag_data[fld]["grid"]["spacing"] = grid_spacing
        diag_data[fld]["grid"]["labels"] = grid_labels
        diag_data[fld]["grid"]["global_offset"] = grid_global_offset
        diag_data[fld]["attributes"] = attrs
    diag_data["field_solver"] = fld_solver
    diag_data["field_solver_params"] = fld_solver_params
    diag_data["field_boundary"] = fld_boundary
    diag_data["field_boundary_params"] = fld_boundary_params
    diag_data["particle_boundary"] = part_boundary
    diag_data["particle_boundary_params"] = part_boundary_params
    diag_data["current_smoothing"] = current_smoothing
    diag_data["charge_correction"] = charge_correction
    return diag_data


def radial_gradient(fld, dr):
    """
    Calculate the radial gradient of a 2D r-z field.

    To obtain an accurate derivative on axis, a wider array which contains
    the initial field and its mirrored view along the axis is created. The
    gradient of this array is computed and only its upper half is returned.

    Parameters
    ----------
    fld : ndarray
        A 2D array containing the original r-z field.
    dr : float
        Radial separation between grid points.

    """
    n_r = fld.shape[1]
    fld_with_mirror = np.concatenate((fld[:, ::-1], fld), axis=1)
    return np.gradient(fld_with_mirror, dr, axis=1)[:, n_r:]



import os

import numpy as np
import matplotlib.pyplot as plt
from wake_t import PlasmaStage, GaussianPulse
from wake_t.utilities.bunch_generation import get_matched_bunch
from openpmd_viewer.addons import LpaDiagnostics
from wake_t.physics_models.laser.laser_pulse import OpenPMDPulse


tests_output_folder = "./tests_output"


def test_openpmd_viewer():
    """
    This test checks that the data generated with Wake-T can be read and
    analyzed with the openPMD viewer. Both the `h5py` and `openpmd-api`
    backends are tested.
    """
    # Diagnostics directory.
    test_dir = os.path.join(tests_output_folder, "openpmd_viewer_test")
    diag_dir = os.path.join(test_dir, "diags")

    # Create laser driver.
    laser = GaussianPulse(100e-6, l_0=800e-9, w_0=50e-6, a_0=3, tau=30e-15, z_foc=0.0)

    # Create bunch (matched to a blowout at a density of 10^{23} m^{-3}).
    en = 1e-6  # m
    ene = 200  # units of beta*gamma
    ene_sp = 0.3  # %
    xi_c = laser.xi_c - 55e-6  # m
    s_t = 10  # fs
    q_tot = 100  # pC
    n_part = 1e4
    np.random.seed(42)
    bunch = get_matched_bunch(en, en, ene, ene_sp, s_t, xi_c, q_tot, n_part, n_p=1e23)

    # Create plasma stage.
    plasma = PlasmaStage(
        1e-2,
        1e23,
        laser=laser,
        wakefield_model="quasistatic_2d",
        n_out=50,
        laser_evolution=True,
        r_max=200e-6,
        r_max_plasma=120e-6,
        xi_min=30e-6,
        xi_max=120e-6,
        n_r=200,
        n_xi=180,
        dz_fields=0.5e-3,
        ppc=5,
    )

    # Do tracking.
    plasma.track(bunch, opmd_diag=True, diag_dir=diag_dir)

    # Check that the data can be analyzed.
    for backend in ["h5py", "openpmd-api"]:
        diags = LpaDiagnostics(os.path.join(diag_dir, "hdf5"), backend=backend)
        its = diags.iterations
        emitt_x = np.zeros(len(its))
        emitt_y = np.zeros(len(its))
        for i, it in enumerate(its):
            diags.get_field("rho", iteration=it, plot=True)
            plt.savefig(os.path.join(diag_dir, "{}_rho_{}.png".format(backend, it)))
            plt.clf()

            diags.get_field("E", "x", iteration=it, plot=True)
            plt.savefig(os.path.join(diag_dir, "{}_Ex_{}.png".format(backend, it)))
            plt.clf()

            emitt_x[i], emitt_y[i] = diags.get_emittance(iteration=it)

        plt.plot(emitt_x)
        plt.plot(emitt_y)
        plt.savefig(os.path.join(diag_dir, "{}_emitt.png".format(backend)))


def test_openpmd_pulse():
    """
    This test checks that the laser envelope generated with Wake-T can be read
    from openPMD files using the OpenPMDPulse class, and that the pulse can be
    used in a Wake-T simulation. The simulation run using this approach is compared
    to a reference simulation where the laser envelope is generated directly with
    Wake-T. Both simulations should yield the same results.
    """
    # Diagnostics directory of the reference simulation
    test_dir = os.path.join(tests_output_folder, "openpmd_viewer_test")
    diag_dir_orig = os.path.join(test_dir, "diags")

    # Load laser envelope from openPMD file
    file_name = os.path.join(diag_dir_orig, "hdf5", "data%T.h5")
    laser = OpenPMDPulse(file_name=file_name, envelope_name="a", iteration=0)

    # Create bunch (matched to a blowout at a density of 10^{23} m^{-3}).
    en = 1e-6  # m
    ene = 200  # units of beta*gamma
    ene_sp = 0.3  # %
    xi_c = 100e-6 - 55e-6  # m
    s_t = 10  # fs
    q_tot = 100  # pC
    n_part = 1e4
    np.random.seed(42)
    bunch = get_matched_bunch(en, en, ene, ene_sp, s_t, xi_c, q_tot, n_part, n_p=1e23)

    # Create plasma stage.
    plasma = PlasmaStage(
        1e-2,
        1e23,
        laser=laser,
        wakefield_model="quasistatic_2d",
        n_out=50,
        laser_evolution=True,
        r_max=200e-6,
        r_max_plasma=120e-6,
        xi_min=30e-6,
        xi_max=120e-6,
        n_r=200,
        n_xi=180,
        dz_fields=0.5e-3,
        ppc=5,
    )

    # Diagnostics directory.
    test_dir = os.path.join(tests_output_folder, "openpmd_pulse_test")
    diag_dir = os.path.join(test_dir, "diags")

    # Do tracking.
    plasma.track(bunch, opmd_diag=True, diag_dir=diag_dir)

    # Analyze data
    backend = "h5py"
    diags = LpaDiagnostics(os.path.join(diag_dir, "hdf5"), backend=backend)
    its = diags.iterations
    emitt_x = np.zeros(len(its))
    emitt_y = np.zeros(len(its))
    for i, it in enumerate(its):
        diags.get_field("rho", iteration=it, plot=True)
        plt.savefig(os.path.join(diag_dir, "{}_rho_{}.png".format(backend, it)))
        plt.clf()

        diags.get_field("E", "x", iteration=it, plot=True)
        plt.savefig(os.path.join(diag_dir, "{}_Ex_{}.png".format(backend, it)))
        plt.clf()

        emitt_x[i], emitt_y[i] = diags.get_emittance(iteration=it)

    plt.plot(emitt_x)
    plt.plot(emitt_y)
    plt.savefig(os.path.join(diag_dir, "{}_emitt.png".format(backend)))

    # Read data from reference simulation
    diags_orig = LpaDiagnostics(os.path.join(diag_dir_orig, "hdf5"), backend=backend)
    its = diags_orig.iterations
    emitt_x_orig = np.zeros(len(its))
    emitt_y_orig = np.zeros(len(its))
    for i, it in enumerate(its):
        emitt_x_orig[i], emitt_y_orig[i] = diags_orig.get_emittance(iteration=it)

    # Check that both simulations yield the same results for the emittance within a one per mil relative tolerance
    np.testing.assert_allclose(emitt_x, emitt_x_orig, rtol=1e-3)
    np.testing.assert_allclose(emitt_y, emitt_y_orig, rtol=1e-3)


if __name__ == "__main__":
    test_openpmd_viewer()
    test_openpmd_pulse()



"""Defines a linearly-varying (in radius) azimuthal magnetic field"""

from wake_t.fields.analytical_field import AnalyticalField
from wake_t.utilities.numba import prange


def b_x(x, y, z, t, bx, constants):
    """B_x component."""
    k = constants[0]
    for i in prange(x.shape[0]):
        bx[i] -= k * y[i]


def b_y(x, y, z, t, by, constants):
    """B_y component."""
    k = constants[0]
    for i in prange(x.shape[0]):
        by[i] += k * x[i]


class LinearBThetaField(AnalyticalField):
    """Defines a linear azimuthal magnetic field.

    Defining `k` as the constant focusing gradient, the field can be expressed
    as:

        b_theta = k * r

    where r is the radius.

    In cartesian coordinates, this is equivalent to:

        b_x = - k * y
        b_y = k * x

    Parameters
    ----------
    foc_gradient : float
        Uniform focusing gradient in T/m.
    """

    def __init__(self, foc_gradient):
        super().__init__(b_x=b_x, b_y=b_y, constants=[foc_gradient])



"""This module contains functions for manupulating bunch distributions"""

import numpy as np


def convert_to_ocelot_matrix(bunch_matrix, w, gamma_ref=None):
    """
    Produces a matrix with the phase space coordinates
    (x, x', y, y', xi, dp) from a matrix containing (x, px, y, py, xi, pz).
    """
    x = bunch_matrix[0]
    y = bunch_matrix[2]
    xi = bunch_matrix[4]
    px = bunch_matrix[1]
    py = bunch_matrix[3]
    pz = bunch_matrix[5]
    g = np.sqrt(1 + px**2 + py**2 + pz**2)
    if gamma_ref is None:
        gamma_ref = np.average(g, weights=w)
    b_ref = np.sqrt(1 - gamma_ref ** (-2))
    dp = (g - gamma_ref) / (gamma_ref * b_ref)
    p_kin = np.sqrt(gamma_ref**2 - 1)
    return np.array([x, px / p_kin, y, py / p_kin, -xi, dp]), gamma_ref


def convert_from_ocelot_matrix(beam_matrix, gamma_ref):
    """
    Produces a matrix with the phase space coordinates
    (x, px, y, py, xi, pz) from a matrix containing (x, x', y, y', xi, dp).

    Parameters
    ----------
    bunch_matrix : array
        6 x N matrix, where N is the number of particles, containing the
        phase-space information of the bunch as (x, x', y, y', xi, dp) in
        units of (m, rad, m, rad, m, -). dp is defined as
        dp = (g-g_ref)/g_ref, while x' = px/p_kin and y' = py/p_kin, where
        p_kin is the kinetic momentum of each particle.
    gamma_ref : float
        Reference energy with respect to which the particle momentum dp is
        calculated.

    """
    dp = beam_matrix[5]
    b_ref = np.sqrt(1 - gamma_ref ** (-2))
    gamma = dp * gamma_ref * b_ref + gamma_ref
    p_kin = np.sqrt(gamma_ref**2 - 1)
    x = beam_matrix[0]
    px = beam_matrix[1] * p_kin
    y = beam_matrix[2]
    py = beam_matrix[3] * p_kin
    xi = -beam_matrix[4]
    pz = np.sqrt(gamma**2 - px**2 - py**2 - 1)
    return np.array([x, px, y, py, xi, pz])


def rotation_matrix_xz(angle):
    """Returns matrix to rotate the beam in the x-z plane"""
    cs = np.cos(angle)
    sn = np.sin(angle)
    return np.array(
        [
            [cs, 0.0, 0.0, 0.0, sn, 0.0],
            [0.0, cs, 0.0, 0.0, 0.0, sn],
            [0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [-sn, 0.0, 0.0, 0.0, cs, 0.0],
            [0.0, -sn, 0.0, 0.0, 0.0, cs],
        ]
    )



"""This module contains methods for saving particle distributions to files"""

from typing import Optional

from aptools.particle_distributions import save_distribution, ParticleDistribution

from wake_t.particles.particle_bunch import ParticleBunch


def save_bunch_to_file(
    bunch: ParticleBunch,
    data_format: str,
    file_path: str,
    species_name: Optional[str] = None,
):
    """Save a particle bunch to file.

    Parameters
    ----------
    bunch : ParticleBunch
        The bunch to save.
    data_format : str
        The output data format. Can be 'astra', 'csrtrack', or 'openpmd'.
    folder_path : str
        Path to the folder where the bunch file will be created.
    file_name : str
        Name of the file to be saved, without format extension.
    species_name : str, optional
        Only required if `data_format='openpmd'`. Name of the particle species
        under which the bunch data will be stored. If None, the name of the
        bunch will be used.
    """
    kwargs = {}

    # For openpmd output, save with species name.
    if data_format == "openpmd":
        if species_name is None:
            species_name = bunch.name
        kwargs["species_name"] = species_name
    # Create APtools distribution and save it.
    distribution = ParticleDistribution(
        x=bunch.x,
        y=bunch.y,
        z=bunch.xi,
        px=bunch.px,
        py=bunch.py,
        pz=bunch.pz,
        w=bunch.w,
        q_species=bunch.q_species,
        m_species=bunch.m_species,
    )
    save_distribution(distribution, file_path, data_format, **kwargs)



"""
This module contains the envelope solver. This module is strongly based on the
paper 'An accurate and efficient laser-envelope solver for the modeling of
laser-plasma accelerators', written by C Benedetti et al in 2018.

Authors: Wilbert den Hertog, Ángel Ferran Pousa, Carlo Benedetti
"""

import numpy as np
import scipy.constants as ct

from wake_t.utilities.numba import njit_serial
from .tdma import TDMA
from .utils import unwrap


@njit_serial(fastmath=True)
def evolve_envelope(
    a, a_old, chi, k0, kp, zmin, zmax, nz, rmax, nr, dt, nt, use_phase=True
):
    """
    Solve the 2D envelope equation
    (\nabla_tr^2+2i*k0/kp*d/dt+2*d^2/(dzdt)-d^2/dt^2)â = chi*â

    Parameters
    ----------
    a0 : array
        Initial value for â at tau=0. Dimension: nz*nr.
    aold : array
        Initial value for â at tau=-1. Dimension: nz*nr.
    chi : array
        Arrays of the values of the susceptibility. Dimension nz*nr.
    k0 : float
        Laser wave number.
    kp : float
        Plasma skin depth.
    zmin : float
        Minimum value for zeta.
    zmax : float
        Maximum value for zeta.
    nz : int
        Number of grid points in zeta-direction.
    rmax : float
        Maximum value for rho (minimum value is always 0).
    nr : int
        Amount of points in the rho direction.
    dt : float
        Tau step size.
    nt : int
        Number of tau steps.
    use_phase : bool
        Determines whether to take into account the terms related to the
        longitudinal derivative of the complex phase.

    """
    # Preallocate arrays. a_old and a include 2 ghost cells in the z direction.
    rhs = np.empty(nr, dtype=np.complex128)

    # Calculate step sizes.
    dz = (zmax - zmin) * kp / (nz - 1)
    dr = rmax * kp / nr
    dt = dt * ct.c * kp

    # Precalculate common fractions.
    inv_dt = 1 / dt
    inv_dr = 1 / dr
    inv_dz = 1 / dz
    inv_dzdt = inv_dt * inv_dz
    k0_over_kp = k0 / kp

    # Initialize phase difference
    d_theta1 = 0.0
    d_theta2 = 0.0

    # Calculate C^+ and C^- [Eq. (8)].
    C_minus = (
        -2.0 * inv_dr**2.0 * 0.5
        - 1j * k0_over_kp * inv_dt
        + 1.5 * inv_dzdt
        - inv_dt**2.0
    )
    C_plus = (
        -2.0 * inv_dr**2.0 * 0.5
        + 1j * k0_over_kp * inv_dt
        - 1.5 * inv_dzdt
        - inv_dt**2.0
    )

    # Calculate L^+ and L^-. Change wrt Benedetti - 2018: in Wake-T we use
    # cell-centered nodes in the radial direction.
    L_base = 1.0 / (2.0 * (np.arange(nr) + 0.5))
    L_minus_over_2 = (1.0 - L_base) * inv_dr**2.0 * 0.5
    L_plus_over_2 = (1.0 + L_base) * inv_dr**2.0 * 0.5

    # Loop over time iterations.
    for n in range(nt):
        # a_new_jp1 is equivalent to a_new[j+1] and a_new_jp2 to a_new[j+2].
        a_new_jp1 = np.zeros(nr, dtype=np.complex128)
        a_new_jp2 = np.zeros(nr, dtype=np.complex128)

        # Getting the phase of the envelope on axis.
        if use_phase:
            phases = unwrap(np.angle(a[:, 0]))

        # Loop over z.
        for j in range(nz - 1, -1, -1):
            # Calculate phase differences between adjacent points.
            if use_phase:
                d_theta1 = phases[j + 1] - phases[j]
                d_theta2 = phases[j + 2] - phases[j + 1]

            # Calculate D factor [Eq. (6)].
            D_jkn = (1.5 * d_theta1 - 0.5 * d_theta2) * inv_dz

            # Calculate right-hand side of Eq (7).
            for k in range(nr):
                rhs_k = (
                    -2 * inv_dt**2 * a[j, k]
                    - ((C_minus - chi[j, k] * 0.5 - 1j * inv_dt * D_jkn) * a_old[j, k])
                    - (
                        2
                        * np.exp(-1j * d_theta1)
                        * inv_dzdt
                        * (a_new_jp1[k] - a_old[j + 1, k])
                    )
                    + (
                        0.5
                        * np.exp(-1j * (d_theta2 + d_theta1))
                        * inv_dzdt
                        * (a_new_jp2[k] - a_old[j + 2, k])
                    )
                )
                if k > 0:
                    rhs_k -= L_minus_over_2[k] * a_old[j, k - 1]
                if k + 1 < nr:
                    rhs_k -= L_plus_over_2[k] * a_old[j, k + 1]
                rhs[k] = rhs_k

            # Calculate diagonals.
            d_main = C_plus - chi[j] * 0.5 + 1j * inv_dt * D_jkn
            d_upper = L_plus_over_2[: nr - 1]
            d_lower = L_minus_over_2[1:nr]

            # Update a_old and a at j+2 with the current values of a a_new.
            a_old[j + 2] = a[j + 2]
            a[j + 2] = a_new_jp2

            # Shift a_new[j+1] to a_new[j+2] in preparation for the next
            # iteration.
            a_new_jp2[:] = a_new_jp1

            # Compute a_new at the current j using the TDMA method and store
            # result in a_new_jp1 to use it in the next iteration.
            TDMA(d_lower, d_main, d_upper, rhs, a_new_jp1)

        # When the left of the computational domain is reached, paste the last
        # few values in the a_old and a arrays.
        a_old[0:2] = a[0:2]
        a[0] = a_new_jp1
        a[1] = a_new_jp2



"""
This module contains the definition of the CSRCalculator class.

As indicated below, the methods for computing the reference trajectory and
the CSR effects on the bunch are strongly based on the 1D CSR model from OCELOT
(https://github.com/ocelot-collab/ocelot) written by S. Tomin and M. Dohlus.

"""

from typing import Optional

import numpy as np
from scipy import stats
import scipy.constants as ct


class CSRCalculator:
    """Class taking care of calculating and applying CSR effects."""

    def __init__(self):
        self._ref_traj = None
        self._lattice_elements = []
        self._lattice_element_steps = []
        self._lattice_element_traj_steps = []
        self.set_settings()

    def set_settings(self, csr_step=0.1, csr_traj_step=0.0005, n_bins=2000):
        """Set the setting for CSR calculation."""
        self._csr_step = csr_step
        self._csr_traj_step = csr_traj_step
        self._n_bins = n_bins

    def add_lattice_element(self, element):
        """Add lattice element to CSR calculator"""
        self._lattice_elements.append(element)
        self._set_element_steps(element)
        self._calculate_trajectory(element)

    def get_csr_step(self, element):
        """Return the CSR step size used in the specified lattice element."""
        el_idx = self._lattice_elements.index(element)
        ds_csr, ds_traj = self._lattice_element_traj_steps[el_idx]
        return ds_csr

    def clear(self):
        """
        Clear the CSR trajectory and list of lattice elements.

        Calling this method allows for a new CSR calculation to be started
        without being affected by the previous beamline.

        """
        self._ref_traj = None
        self._lattice_elements = []
        self._lattice_element_steps = []

    def apply_csr(self, bunch_matrix, bunch_q, gamma, element, element_pos):
        """
        Apply the CSR kick to the specified particle bunch at a certain
        lattice location.

        The implementation of this method is an adaptation of code from OCELOT
        (https://github.com/ocelot-collab/ocelot) written by S. Tomin and
        M. Dohlus.

        Parameters
        ----------
        bunch_matrix : ndarray
            Matrix containing the (x, xp, y, yp, tau, dp) components of each
            bunch particle.
        bunch_q : ndarray
            Array containing the charge of each particle.
        gamma : float
            Reference energy used for the CSR calculation.
        element : TMElement
            Lattice element through which the bunch is currently being tracked.
        element_pos : float
            Current position (longitudinal) of the bunch in the lattice
            element.

        """
        el_idx = self._lattice_elements.index(element)
        ds_csr, ds_traj = self._lattice_element_traj_steps[el_idx]

        # Get current position along trajectory.
        s_current = 0
        for el in self._lattice_elements:
            if el is element:
                break
            s_current += el.length
        s_current += element_pos

        # Get line charge profile of bunch as a histogram.
        z = -bunch_matrix[4]
        bunch_hist, bin_edges = np.histogram(z, self._n_bins, weights=bunch_q)

        # Make profile smooth (important to prevent instabilities) using
        # a Gaussian KDE. This method might be changed/optimized.
        bin_size = bin_edges[1] - bin_edges[0]
        n_part = len(bunch_q)
        bin_centers = bin_edges[1:] - bin_size / 2
        bunch_pdf = stats.gaussian_kde(
            bin_centers, bw_method=n_part ** (-1 / 5), weights=bunch_hist
        )
        bunch_hist = bunch_pdf(bin_centers) * bin_size * bunch_hist.sum()
        bin_center_0 = bin_centers[0]

        # Determine iteration range.
        s_array = self._ref_traj[0, :]
        idx = (np.abs(s_array - s_current)).argmin()
        idx_prev = (np.abs(s_array - (s_current - ds_csr))).argmin()
        it_range = np.arange(idx_prev, idx) + 1
        n_iter = len(it_range)

        # Calculate CSR kernel.
        K1 = 0
        for it in it_range:
            K1 += self._calculate_kernel(
                it, self._ref_traj, self._n_bins, bin_size, gamma
            )
        K1 /= n_iter

        # Convolve kernel with line charge
        lam_K1 = np.convolve(bunch_hist, K1[::-1]) / bin_size * ds_csr

        # Calculate and apply energy kick
        z_norm = z * (1.0 / bin_size) - bin_center_0 / bin_size
        dE = np.interp(z_norm, np.arange(len(lam_K1)), lam_K1)
        pc_ref = np.sqrt(gamma**2 - 1) * 0.511e-3
        delta_p = dE * 1e-9 / pc_ref
        bunch_matrix[5] += delta_p

    def _set_element_steps(self, element):
        """
        Calculate and store the csr and trajectory step size for the
        given element.

        """
        n_out = element.n_out
        ds_csr = self._csr_step
        ds_traj = self._csr_traj_step
        if n_out is not None:
            ds_out = element.length / n_out
            ds_csr = ds_out / np.ceil(ds_out / ds_csr)
        elif element.length < ds_csr:
            ds_csr = element.length
        elif not element.length % ds_csr == 0:
            ds_csr = element.length / np.ceil(element.length / ds_csr)
        ds_traj = ds_csr / np.ceil(ds_csr / ds_traj)
        self._lattice_element_traj_steps.append([ds_csr, ds_traj])

    def _calculate_trajectory(self, element):
        """
        Calculate the reference trajectory along the given element.

        The implementation of this method is an adaptation of code from OCELOT
        (https://github.com/ocelot-collab/ocelot) written by S. Tomin and
        M. Dohlus.

        """
        el_idx = self._lattice_elements.index(element)
        ds_csr, ds_traj = self._lattice_element_traj_steps[el_idx]
        n_steps = int(np.ceil(element.length / ds_traj))
        ds_traj = element.length / n_steps
        traj = np.zeros((7, n_steps))
        if self._ref_traj is None:
            self._ref_traj = np.transpose([[0, 0, 0, 0, 0, 0, 1.0]])
        traj_start = self._ref_traj[:, [-1]]
        e1 = traj_start[4:]
        l_traj = np.linspace(ds_traj, element.length, n_steps)
        traj[0, :] = traj_start[0] + l_traj
        if element.theta != 0:
            rho = element.length / element.theta
            # to support element tilt in the future
            tilt = 0
            rho_x = rho * np.cos(tilt)
            rho_y = rho * np.sin(tilt)
            rho_vect = np.array([-rho_y, rho_x, 0])

            n_vect = rho_vect / rho
            e2 = np.cross(n_vect, e1.T).T
            si = np.sin(l_traj / rho)
            co = np.cos(l_traj / rho)
            omco = 2 * np.sin(l_traj / (2 * rho)) ** 2

            traj[1:4, :] = traj_start[1:4] + rho * (e1 * si + e2 * omco)
            traj[4:, :] = e1 * co + e2 * si

        else:
            traj[1:4, :] = traj_start[1:4] + e1 * l_traj
            traj[4:, :] = e1

        self._ref_traj = np.append(self._ref_traj, traj, axis=1)
        self._lattice_element_steps.append(n_steps)

    def _calculate_kernel(self, i, traj, n_bins, bin_size, gamma):
        """
        Calculate CSR kernel.

        The implementation of this method is an adaptation of code from OCELOT
        (https://github.com/ocelot-collab/ocelot) written by S. Tomin and
        M. Dohlus.

        Parameters
        ----------
        i : int
            Iteration index.
        traj : ndarray
            Reference trajectory along which CSR forces are calculated
        n_bins : int
            Number of bins of the longitudinal bunch histogram.
        bin_size : int
            Size of the histogram bins.
        gamma: float
            reference gamma to calculate CSR kernel.

        """

        w_range = np.arange(-n_bins, 0) * bin_size

        # Calculate long-range interactions.
        w, KS = self._calculate_kernel_long_range(i, traj, w_range[0], gamma)

        # Calculate short-range interactions.
        w_min = np.min(w)
        if w_range[0] < w_min:
            m = np.where(w_range < w_min)[0][-1]
            KS2 = self._calculate_kernel_short_range(
                i, traj, np.append(w_range[0 : m + 1], w_min), gamma
            )
            KS1 = KS[0]
            KS2 = (KS2[-1] - KS2) + KS1
            KS = np.append(KS2[0:-1], np.interp(w_range[m + 1 :], w, KS))
        else:
            KS = np.interp(w_range, w, KS)

        four_pi_eps0 = 1.0 / (1e-7 * ct.c**2)
        K1 = np.diff(np.diff(KS, append=0), append=0) / bin_size / four_pi_eps0
        return K1

    def _calculate_kernel_long_range(self, i, traj, wmin, gamma):
        """
        The implementation of this method is an adaptation of code from OCELOT
        (https://github.com/ocelot-collab/ocelot) written by S. Tomin and
        M. Dohlus.

        For details about the CSR model, see section 2.6 of:
        https://www.desy.de/~dohlus/UWake/
        Two%20Methods%20for%20the%20Calculation%20of%20CSR%20Fields.pdf

        Parameters
        ----------
        i : int
            Iteration index.
        traj: ndarray
            Reference trajectory along which CSR forces are calculated.
        wmin: float
            Leftmost edge of the longitudinal bunch histogram.
        gamma: float
            reference gamma to calculate CSR kernel.

        """

        # Relativistic parameters
        gamma_sq_inv = 1.0 / gamma**2
        beta_sq = 1.0 - gamma_sq_inv
        beta = np.sqrt(beta_sq)

        i_0 = self._estimate_start_index(i, traj, wmin, beta)
        # i_0 = 0

        # Reference trajectory positions with respect to the current one.
        s = traj[0, i_0:i] - traj[0, i]

        # Distance (in x, y, z) from the current to previous trajectory points.
        n0 = traj[1, i] - traj[1, i_0:i]
        n1 = traj[2, i] - traj[2, i_0:i]
        n2 = traj[3, i] - traj[3, i_0:i]

        # Distance from the current to previous trajectory points.
        R = np.sqrt(n0 * n0 + n1 * n1 + n2 * n2)

        w = s + beta * R

        # Indices of trajectory locations where w<wmin
        j = np.where(w <= wmin)[0]
        if len(j) > 0:
            j = j[-1]
            w = w[j:i]
            s = s[j:i]
            n0 = n0[j:]
            n1 = n1[j:]
            n2 = n2[j:]
            R = R[j:]
            j += i_0
        else:
            j = i_0

        # Tangential unit vectors along trajectory until current position.
        t4 = traj[4, j:i]
        t5 = traj[5, j:i]
        t6 = traj[6, j:i]

        t4_i = traj[4, i]
        t5_i = traj[5, i]
        t6_i = traj[6, i]

        # Calculate kernel
        R_inv = 1 / R
        n0 *= R_inv
        n1 *= R_inv
        n2 *= R_inv

        x = n0 * t4 + n1 * t5 + n2 * t6
        K = (
            beta * (x - n0 * t4_i - n1 * t5_i - n2 * t6_i)
            - beta_sq * (1.0 - t4 * t4_i - t5 * t5_i - t6 * t6_i)
            - gamma_sq_inv
        ) * R_inv - (1.0 - beta * x) / w * gamma_sq_inv

        # Integrate kernel
        K[:-1] += K[1:]
        K *= 0.5 * np.diff(s, append=0)
        K = np.cumsum(K[::-1])[::-1]

        return w, K

    def _calculate_kernel_short_range(self, i, traj, w_range, gamma):
        """
        Calculate short-range contributions to CSR kernel.

        The implementation of this method is an adaptation of code from OCELOT
        (https://github.com/ocelot-collab/ocelot) written by S. Tomin and
        M. Dohlus.

        Parameters
        ----------
        i : int
            Iteration index.
        traj: ndarray
            Reference trajectory along which CSR forces are calculated.
        wrange: ndarray
            Region of beam < w_min.
        gamma: float
            reference gamma to calculate CSR kernel.

        """
        # Relativistic parameters
        gamma_sq = gamma**2
        gamma_sq_inv = 1.0 / gamma_sq
        beta_sq = 1.0 - gamma_sq_inv
        beta = np.sqrt(beta_sq)

        # winf
        Rv1 = traj[1:4, i] - traj[1:4, 0]
        s1 = traj[0, 0] - traj[0, i]
        ev1 = traj[4:, 0]
        evo = traj[4:, i]
        winfms1 = np.dot(Rv1, ev1)

        aup = -Rv1 + winfms1 * ev1
        a2 = np.dot(aup, aup)
        a = np.sqrt(a2)

        uup = aup / a if a != 0 else None

        winf = s1 + winfms1
        s = winf + gamma * (
            gamma * (w_range - winf)
            - beta * np.sqrt(gamma_sq * (w_range - winf) ** 2 + a2)
        )
        R = (w_range - s) / beta

        KS = (
            beta * (1.0 - np.dot(ev1, evo)) * np.log(R[0] / R)
            - (beta_sq * np.dot(ev1, evo) - 1.0)
            * np.log((winf - s + R) / (winf - s[0] + R[0]))
            + gamma_sq_inv * np.log(w_range[0] / w_range)
        )
        if a2 / R[1] ** 2 > 1e-7:
            KS -= (
                beta
                * np.dot(uup, evo)
                * (np.arctan((s[0] - winf) / a) - np.arctan((s - winf) / a))
            )
        return KS

    def _estimate_start_index(self, i, traj, w_min, beta, i_min=1000, n_test=10):
        """
        This method estimates the index of the first trajectory point from
        which CSR effects should be computed.

        This method can significantly reduce the computing time of CSR effects
        by pre-discaring regions of the reference trajectory which do not
        influence the CSR calculation (i.e. the points where w <= wmin). This
        is performed by testing the w <= wmin condition for a subset of n_test
        equally-spaced points along the reference trajectory. The index of the
        last tested trajectory point in which w <= wmin is returned.

        Parameters
        ----------
        i : int
            Iteration index
        traj : ndarray
            Reference trajectory along which CSR forces are calculated.
        w_min : float
            Leftmost edge of the longitudinal bunch binning.
        beta : float
            Relativistic factor.
        i_min : int
            Minimum iteration index. When i<i_min, no estimation of the
            starting index is performed (0 is returned).
        n_test : int
            Number of points along the trajectory in which to test whether they
            should be taken into account for the CSR calculation.

        Returns
        -------
        The estimated start index, which is always <= than the real one.

        """
        i_0 = 0
        if i > i_min:
            idx = np.linspace(0, i, n_test, dtype=np.int32)
            s = traj[0, idx] - traj[0, i]
            n0 = traj[1, i] - traj[1, idx]
            n1 = traj[2, i] - traj[2, idx]
            n2 = traj[3, i] - traj[3, idx]
            R = np.sqrt(n0 * n0 + n1 * n1 + n2 * n2)
            w = s + beta * R
            j = np.where(w <= w_min)[0]
            if len(j) > 0:
                i_0 = idx[j[-1]]
        return i_0


_csr_calculator = CSRCalculator()


def get_csr_calculator() -> CSRCalculator:
    """Return the single instance of CSRCalculator"""
    return _csr_calculator


def set_csr_settings(
    csr_step: Optional[float] = 0.1,
    csr_traj_step: Optional[float] = 0.0005,
    n_bins: Optional[int] = 2000,
) -> None:
    """
    Set the setting for CSR calculation.

    Parameters
    ----------
    csr_step : float
        Iteration index.
    csr_traj_step : float
        Reference trajectory along which CSR forces are calculated.
    n_bins : int
        Number of bins used for determining the longitudinal charge profile
        of the bunch.

    """
    _csr_calculator.set_settings(csr_step, csr_traj_step, n_bins)


def reset_csr_calculator() -> None:
    """
    Reset CSR calculator by clearing the stored reference trajectory and
    lattice elements. Needed to start a new, independent calculation after
    a previous one.

    """
    _csr_calculator.clear()
